#include <Python.h>
#include <casa/aips.h>
#include <casa/Arrays.h>
#include <casa/Containers/Record.h>
#include <casa/Arrays/ArrayPosIter.h>
#include <tables/Tables.h>
#include <tables/Tables/BaseTable.h>
#include <tables/Tables/TableParse.h>
#include <tables/Tables/TableKeyword.h>

#include "stringifier.h"
#include "countedarray.h"

#include <boost/python.hpp>

#include <arrayobject.h>

// stl cruft
#include <map>
#include <algorithm>
#include <sstream>
#include <typeinfo>

// Note:
// We will include the actual numarray API later on.
// First we will define some template stuff which the 
// API stuff is supposed to specialize.
//
#ifndef NUMTYPEIF
    #error You must edit the Makefile and set the NUMTYPE variable!
#endif

#define VERS "$Id: pycasa.cc,v 1.10 2007/09/14 13:46:19 jive_cc Exp $"

//using namespace casa;
using namespace std;
using namespace boost::python;

// exception cruft
struct column_proxy_no_such_column:
    public std::exception 
{};
struct column_proxy_unsupported_column_datatype:
    public std::exception 
{};

struct failed_to_load_array_API:
    public std::exception
{};

struct failed_to_create_Numarray_object:
    public std::exception
{};

struct pyobject_is_notapybool:
    public std::exception
{};

struct incompatible_pythontype_for_columndatatype:
    public std::exception
{};

struct incompatible_pythontype_for_requesteddatatype:
    public std::exception
{};

struct shape_of_subarrays_not_all_equal:
    public std::exception
{};

struct method_accepts_only_kwargs:
    public std::exception
{};

struct unsupported_PyArray_sourcetype:
    public std::exception
{};

struct failed_conversion:
    public std::exception
{};
struct unsupported_conversion:
    public std::exception
{};

// some support stuff
// Define it before we include the *_if.h file
// so they may make use of it..
// [eg Numeric does not typedef a complex, it just
//  records a size equal to 2 floats/doubles per
//  arrayelement]
namespace support {
    // PyBool objects don't hold a value
    // but point at the static Py_True or
    // Py_False objects. The thing they point
    // at determines the truthiness of the PyBool.
    // We need a fn which translates this pointer-to-truthiness
    // into a plain old c++ boolean.
    // Throws exception if o does not point to
    // either Py_True||Py_False
    bool pybool_ascppbool(PyObject* o) {
        bool rv;

        if( o==Py_False ) {
            rv = false;
        } else {
            if( o==Py_True ) {
                rv = true;
            } else {
                throw pyobject_is_notapybool();
            }
        }
        return rv;
    }

    // be able to translate between
    // PyComplex and PyDComplex and
    // a two-element-tuple of float/double
    // (which will allow easy translation to
    // other objects (eg casa::Complex...)
    template <typename T>
    struct mcplx :
        public std::complex<T> 
    {
        typedef T basic_type;

        // default: uninitialized...
        mcplx(): std::complex<T>()
        {}

        // construct from 'array of Ts'... array must be
        // at least 2T's long eh?
        mcplx( const T* tptr ):
            std::complex<T>(*tptr++, *tptr)
        {}
        
        // construct from real/imag of type T
        mcplx(T r, T i):
            std::complex<T>(r,i)
        {}
        
        // construct from a Py_complex,
        // doing type-casting
        // Compiler should warn if we're doing
        // something which is well.. less optimal
        mcplx(const Py_complex& pc ):
            std::complex<T>(pc.real, pc.imag)
        {}

        // support creating from anything
        // which support .real()/.imag()
        template <typename U>
        mcplx( const U& u ):
            std::complex<T>( u.real(), u.imag() )
        {}
    };

    // hmmmm python check fn's are #defines
    // and we cannot stick those in templates. Bummer.
    // have to have wrappers. Shoot.
    int pyint_check(PyObject* o) {
        return PyInt_Check(o);
    }
    int pylong_check(PyObject* o) {
        return PyLong_Check(o);
    }
    int pyfloat_check(PyObject* o) {
        return PyFloat_Check(o);
    }
    int pycomplex_check(PyObject* o) {
        return PyComplex_Check(o);
    }
    int pystring_check(PyObject* o) {
        return PyString_Check(o);
    }
    int pydict_check(PyObject* o) {
        return PyDict_Check(o);
    }
    int pybool_check(PyObject* o) {
        return PyBool_Check(o);
    }
    int pyobject_check(PyObject* o) {
        // note: cannot do
        //  return (int)o;
        // since then we may be up the
        // creek on LP64 systems where int
        // is still 32bit and ptr is 64bit...
        return ((o==0)?(0):(1));
    }
}

// this template does nothing - it's supposed to be
// specialized in the *_if.h files.
// If we do not define it here, there's nothing
// to specialize, is there? ;)
template <int>
struct pyarray {};


// include the interface for the actually
// used num* package.
// The 'arrayobject.h' of the package is
// already included.
//
// The num* stuff share the following
// so you do NOT need to define/implement those
//  PyArray_TYPES    enum,
//  PyArray_FromDims fn/macro
//  PyArray_Check    fn/macro
//
// But you MUST define the following functions:
// Note: you may assume that the PyObject is
// already checked to be of PyArrayObject* type
//
// *  PyArray_TYPES pyarraytype(PyObject* o) 
//       Return the type of stuff held in the numarray
//    
// *  int pyarrayndim(PyObject*)
//       Return the number of dimensions in the array
//
// *  int pyarraydim(PyObject*, int n)
//       Return the size of dimension 'n'
//
// *  int pyarrayelementsize(PyObject*)
//       Return the size of individual elements in bytes
// 
// *  char* pyarraybytes(PyObject*)
//       Return a char* pointing at the first byte
//       of the datapart
//       
// *  char* pyarraygetptr1(PyObject*, n)
//       Return a pointer to the first byte of element #n 
//       from the first dimension of the array.
//       {used in recursion}
//
// *  bool pyarrayiscontiguous(PyObject*)
//       Return wether the underlying storage is contiguous?
//
// *  bool pyarrayisobject(PyObject*)
//       Returns wether this is an n-dimensional array
//       of PyObject*'s
//
// * int pyarray_arraycheck(PyObject*)
//       Should do an Exact check, not the 'isSubtype'
//       stuff.
//
// =====>    ALSO you must define the following specializations <======
//
//   For all the elementtypes (out of PyArray_TYPES) you 
//   want to support do the following:
// 
// template <>
// struct pyarray<[PyArray_TYPE]> {
//      typedef [sometype] basic_type;
// };
//
// Where [sometype] is the plain-old-data type that the package
// uses to store the elements indicated by PyArray_TYPE
//
// eg:
//    In numarray we find for the PyArray_INT it uses tInt32.
//    So in numarray_if.h we put the following code:
//
//    template <>
//    struct pyarray<PyArray_INT> {
//          typedef tInt32    basic_type;
//    };
//
// Then we repeat for all other types we support.
// Look eg in the 'dtmap' below to find out which types
// pycasa supports. It may be useful if you supported those
// as well... ;)
//
// Please use the #define below to generate the specializations

// pat = P(y)A(rray_)T(ype)
// pod = P(lain)O(ld)D(atatype)
#define PYCASA_DEFINETYPEMAP(pat,pod) \
template <> \
struct pyarray<(int)pat> { \
    typedef pod   basic_type; \
    static const unsigned int elmnt_size = sizeof(basic_type);\
};
// Sometimes, sizeof(basic_type) is NOT how much bytes/element
// the numeric package actually stores.
// See the 'complex' in Numeric: it uses float|double as
// basic type but records 2*sizeof(basic_type) as element size.
// For those cases you can use the one below and specifiy the 
// size-per-element by hand
#define PYCASA_DEFINETYPEMAPEXT(pat,pod,sz) \
template <> \
struct pyarray<(int)pat> { \
    typedef pod   basic_type; \
    static const unsigned int elmnt_size = sz;\
};

#include NUMTYPEIF


// The typelist idea shamelessly 'stolen' from Loki
struct NullType {};

template <typename T, typename U>
struct TypeList {
    typedef T  Head;
    typedef U  Tail;
};

#define TYPELIST1(T1) TypeList<T1,NullType>
#define TYPELIST2(T1,T2) TypeList<T1,TYPELIST1(T2) >
#define TYPELIST3(T1,T2,T3) TypeList<T1,TYPELIST2(T2,T3) >
#define TYPELIST4(T1,T2,T3,T4) TypeList<T1,TYPELIST3(T2,T3,T4) >
#define TYPELIST5(T1,T2,T3,T4,T5) TypeList<T1,TYPELIST4(T2,T3,T4,T5) >
#define TYPELIST6(T1,T2,T3,T4,T5,T6) TypeList<T1,TYPELIST5(T2,T3,T4,T5,T6) >
#define TYPELIST7(T1,T2,T3,T4,T5,T6,T7) TypeList<T1,TYPELIST6(T2,T3,T4,T5,T6,T7) >
#define TYPELIST8(T1,T2,T3,T4,T5,T6,T7,T8) TypeList<T1,TYPELIST7(T2,T3,T4,T5,T6,T7,T8) >
#define TYPELIST9(T1,T2,T3,T4,T5,T6,T7,T8,T9) TypeList<T1,TYPELIST8(T2,T3,T4,T5,T6,T7,T8,T9) >
#define TYPELIST10(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10) TypeList<T1,TYPELIST9(T2,T3,T4,T5,T6,T7,T8,T9,T10) >


#define TLSTR(tl) \
    tl::Head##TLSTR(tl::Tail)

// define the stuff we want to do with a numarray

//@@@@@@@@@@@@@@@@ Support for num* packages -> casa 
namespace arraysupport {

    // struct podcast - casts Plain-Old-Datatypes into each other
    // generic slow impl
    template <typename T, typename U>
    struct podcast {
        static countedarray<T> do_convert_n(const void* src,
                                            unsigned int nelem,
                                            unsigned int elsize) {
            T*          tmp = new T[nelem];
            const char* srcptr = reinterpret_cast<const char*>(src);

            for( unsigned int i=0; i<nelem; ++i, srcptr+=elsize )
                tmp[i] = T( *((const U*)srcptr) );
            return countedarray<T>(tmp);
        }
    };
    // same type: nothing to do
    template <typename T>
    struct podcast<T,T> {
        static countedarray<T> do_convert_n(const void* src, unsigned int, unsigned int) {
            return countedarray<T>( const_cast<T*>(reinterpret_cast<const T*>(src)), 42 );
        }
    };
   
    // the cmplxcaster - transforms arrays of complex numbers into
    // arrays of complex numbers of a different type.
    // We only require that:
    //   * 'src' points at a 'simple' array of objects of type U
    //     (or can be interpreted as such!)
    //   * the source type 'U' supports .real() and .imag() methods
    //   * the destination type 'T' supports construction from
    //     T(real, imag)
    // I think most implementations of complex numbers support this...
    template <typename T, typename U>
    struct cmplxcast {
        // this is the signature we must support
        static countedarray<T> do_convert_n(const void* src,
                                            unsigned int nelem,
                                            unsigned int elsize) {
            T*          tmp = new T[nelem];
            const char* srcptr = reinterpret_cast<const char*>(src);

            for( unsigned int i=0; i<nelem; ++i, srcptr+=elsize )
                tmp[i] = T( ((const U*)srcptr)->real(),
                            ((const U*)srcptr)->imag() );
            return countedarray<T>(tmp);
        }
    };
    // specialization for same type
    template <typename T>
    struct cmplxcast<T,T> {
        static countedarray<T> do_convert_n(const void* src, unsigned int, unsigned int) {
            return countedarray<T>( const_cast<T*>(reinterpret_cast<const T*>(src)), 42 );
        }
    };
    // and specializations for types we know have identical layout
    // but are known under different names
    template <typename T>
    struct cmplxcast<support::mcplx<T>, std::complex<T> > {
        typedef typename support::mcplx<T>   DT;
        static countedarray<DT> do_convert_n(const void* src, unsigned int, unsigned int) {
            return countedarray<DT>(const_cast<DT*>(reinterpret_cast<const DT*>(src)), 42);
        }
    };
    template <typename T>
    struct cmplxcast<std::complex<T>, support::mcplx<T> > {
        typedef typename std::complex<T>   DT;
        static countedarray<DT> do_convert_n(const void* src, unsigned int,unsigned int) {
            return countedarray<DT>(const_cast<DT*>(reinterpret_cast<const DT*>(src)), 42);
        }
    };

    // and the stringcast :)
    // Only works when the underlying datatype is
    // char* i guess so block usage of it in unknown
    // cases
    template <typename, typename> struct stringcast {};

    // this is the version we can support
    // T must a type which support constructing
    // from (const char*, [unsigned] int)
    // ie pointer + length
    //
    // Actually, we rely on the signed char being same
    // size as plain char, right? ... ieeeuw...
    template <typename T>
    struct stringcast<T,char> {
        static countedarray<T> do_convert_n(const void* src,
                                            unsigned int n,
                                            unsigned int strsize) {
            T*            dst = new T[n];
            const char*   srcptr = reinterpret_cast<const char*>(src);
            for( unsigned int i=0; i<n; ++i ) 
                dst[i] = T(srcptr+(i*strsize), strsize);
            return countedarray<T>( dst );
        }
    };
    template <typename T>
    struct stringcast<T,signed char> {
        static countedarray<T> do_convert_n(const void* src,
                                            unsigned int n,
                                            unsigned int strsize) {
            return stringcast<T,char>::do_convert_n(src, n, strsize);
        }
    };
    template <typename T>
    struct stringcast<T,unsigned char> {
        static countedarray<T> do_convert_n(const void* src,
                                            unsigned int n,
                                            unsigned int strsize) {
            return stringcast<T,char>::do_convert_n(src, n, strsize);
        }
    };

    // and the boolcast
    // by default support nothing
    template <typename, typename> struct boolcast {};

    // and specialize if the destination type is bool
    template <typename T>
    struct boolcast<bool, T> {
        static countedarray<bool> do_convert_n(const void* src,
                                               unsigned int n,
                                               unsigned int elsize) {
            bool*         dst = new bool[n];
            const char*   srcptr = reinterpret_cast<const char*>(src);
            cout << "boolcast!" << endl;
            for( unsigned int i=0; i<n; ++i, srcptr+=elsize )
                dst[i] = (*(const T*)srcptr)?(true):(false);
            return countedarray<bool>( dst );
        }
    };
    
    template <typename T>
    struct TypeConv {
        // converters always return a
        // countedarray-of-T's
        typedef countedarray<T>              return_type;

        // the signature of the conversionfunctions is
        // now fixed
        typedef return_type (*func_type)( const void* src,
                                          unsigned int nelements,
                                          unsigned int elsize );
        // the type of the key based on which
        // the functions are stored
        typedef std::string                   key_type;

        // and they are stored in the following map
        typedef std::map<key_type, func_type> map_type;

        // Default c'tor. Immediately fill in conversion
        // for identity, if not there yet
        TypeConv() {
            if( TypeConv<T>::template find<T>() == 0 ) 
                TypeConv<T>::template define<T>( &podcast<T,T>::do_convert_n );
        }

        // register a conversionfunction which
        // transforms an array of type U to 
        // an array of type T.
        // Throws if failure to insert into map.
        template <typename U>
        static void define(func_type fptr) {
            key_type                          k;
            typename map_type::const_iterator curconv;
          
            // construct the key
            k = TypeConv<T>::template make_key<U>();

            if( (curconv=__converters.find(k))==__converters.end() ) {
                std::pair<typename map_type::iterator,bool>   insres;

                insres = __converters.insert( std::make_pair(k, fptr) );
                if( !insres.second ) {
                    cerr << "Failed to insert functionpointer for conversion "
                         << "from " << typeid(U).name() << " => " << typeid(T).name()
                         << endl;
                    throw unsupported_conversion();
                }
            }
        }

        // find if a conversion is defined
        template <typename U>
        static func_type find( void ) {
            key_type                          k = TypeConv<T>::template make_key<U>();
            func_type                         rv( 0 );
            typename map_type::const_iterator curconv;

            if( (curconv=__converters.find(k))!=__converters.end() ) 
                rv = curconv->second;
            return rv;
        }

        private:
            // create a key from the combination T,U
            template <typename U>
            static key_type make_key( void ) {
                std::ostringstream   s;
                s << typeid(U).name() << ':' << typeid(T).name();
                return s.str();
            }

            // keep the function pointers in here
            static map_type __converters;
    };

    // Nice one, this is -> turn a TypeList (TL)
    // into a series of registrations of cast-function-pointers
    // of all the types in the TypeList to the Destination type (DST)
    //
    // In order to stop the Chain, there is another Chain<> template
    // defined below [not-quite-a specialization] which will evaluate to
    // something of a no-op.
    //
    // Takes a template-template parameter which defaults to podcast.
    // It must be a class taking two templateparameters (To,From) and
    // feature a 'do_convert_n' function which is supposed to
    // transform an array of From's into an array of To's.
    template <typename TL, typename DST,
              template <typename,typename> class CASTER = podcast>
    struct Chain {
        // typedefs for shorthand. And for defining the next
        // element in the chain
        typedef typename TL::Head                      SRC;
        typedef TypeConv<DST>                          typeconv_type;
        typedef Chain<typename TL::Tail, DST, CASTER>  next_type;

        // The default c'tor which will initialize our statics. If
        // we do this + declare a static instance of _this_ type, that
        // apparently is enough for triggering instantiation/initialization
        // of all the other entries in the chain! weehee!
        // Note: we must make sure it gets called though ;)
        Chain() {
            typeconv_type   tc;
            
            // Register the function pointer that
            // points at the [static...] 'CASTER<DST,SRC>::do_convert_n'
            // function. Will throw if failure to register
            tc.template define<SRC>( &CASTER<DST,SRC>::do_convert_n );
#if 0
            cout << "Registered conversion of "
                 << typeid(SRC).name() << " -> " << typeid(DST).name()
                 << endl;
#endif
            // trigger constructor of next element in the chain
            __next__instance     = next_type();
        }

        // Declare statics of the 'next_type' 
        // the constructor(s) will take care of what needs to be done
        static next_type              __next__instance;
    };

    // if tail of the TypeList==NullType, this means
    // end of the typelist
    template <typename T, template <typename, typename> class C>
    struct Chain<NullType, T, C> {
        // Only have to define these so the names
        // exist
        typedef NullType      typeconv_type;
        typedef NullType      next_type;
    };

#define BaseType(a) pyarray<a>::basic_type
#define ELSIZE(a) pyarray<a>::elmnt_size

#define TYPEANDNUM(a) a << "(" << #a << ")"

#define SHOWCONV(a,b) \
    "from " << typeid(a).name() << " to " << typeid(b).name() << endl

    // Set up conversion to D via possible warper AT
    //
    // D  = destination type
    // AT = 'as-type', ie as_float, as_int (see below)
    //
    // ie the intermediate type in the conversion chain
    // PyObject* -> (intermediate) -> casa datatype
    // (D should be a casa:: datatype (eg casa::Int)
    //
    // Note: we set up the conversion for
    // a specific object as the conversion parameters
    // are dependant on whatever's stored *in* the object.
    template <typename D, typename AT>
    struct conv {
        // construct for a give PyArrayObject.
        // Assumes that o IS-A PyArrayObject, you're
        // responsible for making sure it is.
        conv( PyObject* o ):
            __cnv( 0 ),
            __ptr( o )
        {
            std::ostringstream  cnvtxt;
            // Set up the conversion thingy
            // inspect what's in the PyArray thingy
            switch( pyarraytype(o) ) {
                case PyArray_CHAR:
                    cnvtxt << SHOWCONV(BaseType(PyArray_CHAR), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_CHAR)>();
                    // arrays of strings are arrays of chars with elementsize
                    // set to the length of the individual strings i guess
                    __esize = pyarrayelementsize(o);
                    break;
#if HAVE_PYARRAYSBYTE
                case PyArray_SBYTE:
                    cnvtxt << SHOWCONV(BaseType(PyArray_SBYTE), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_SBYTE)>();
                    // arrays of strings are arrays of chars with elementsize
                    // set to the length of the individual strings i guess
                    __esize = pyarrayelementsize(o);
                    break;
#endif
                case PyArray_INT:
                    cnvtxt << SHOWCONV(BaseType(PyArray_INT), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_INT)>();
                    __esize = ELSIZE(PyArray_INT);
                    break;
                case PyArray_UINT:
                    cnvtxt << SHOWCONV(BaseType(PyArray_UINT), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_UINT)>();
                    __esize = ELSIZE(PyArray_UINT);
                    break;
                case PyArray_DOUBLE:
                    cnvtxt << SHOWCONV(BaseType(PyArray_DOUBLE), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_DOUBLE)>();
                    __esize = ELSIZE(PyArray_DOUBLE);
                    break;
                case PyArray_FLOAT:
                    cnvtxt << SHOWCONV(BaseType(PyArray_FLOAT), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_FLOAT)>();
                    __esize = ELSIZE(PyArray_FLOAT);
                    break;
                case PyArray_CFLOAT:
                    cnvtxt << SHOWCONV(BaseType(PyArray_CFLOAT), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_CFLOAT)>();
                    __esize = ELSIZE(PyArray_CFLOAT);
                    break;
                case PyArray_CDOUBLE:
                    cnvtxt << SHOWCONV(BaseType(PyArray_CDOUBLE), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_CDOUBLE)>();
                    __esize = ELSIZE(PyArray_CDOUBLE);
                    break;
// some num* implementations define _INT and _LONG
// to be equal under certain conditions (eg 32bit, or,
// 'not LP64')
#if HAVE_PYARRAYLONG
                case PyArray_LONG:
                    cnvtxt << SHOWCONV(BaseType(PyArray_LONG), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_LONG)>();
                    __esize = ELSIZE(PyArray_INT);
                    break;
#endif
#if HAVE_PYARRAYSTRING
                case PyArray_STRING:
                    cnvtxt << SHOWCONV(BaseType(PyArray_STRING), D);
                    __cnv   = TypeConv<D>::template find<BaseType(PyArray_STRING)>();
                    __esize = pyarrayelementsize(o);
                    break;
#endif
                default:
                    cnvtxt << "Unsupported PyArray sourcetype #"
                           << pyarraytype(o);
                    PyErr_SetString(PyExc_ValueError, cnvtxt.str().c_str());
                    throw boost::python::error_already_set();
            }
            if( !__cnv ) {
                string txt;
                txt = "Unsupported conversion " + cnvtxt.str();
                PyErr_SetString(PyExc_ValueError, txt.c_str());
                throw boost::python::error_already_set();
            }
        }

        // this may return the array-data itself or
        // a temporary copy. Just you don't worry 'bout it
        // anyway 'cuz its counted an' da system makes
        // sure that only the temp stuff gets deleted
        // but if it references an existing array that
        // won't be delete[]'d. Hoorah.
        countedarray<D> operator()( void ) const {
            void*  arraydata( pyarraybytes(__ptr) );
            return __cnv( arraydata,
                          PyArray_SIZE((PyArray(__ptr))),
                          __esize );
        }

        private:
            // conversion fn ptr. Actual type is taken
            // from the TypeConv(ersion) 'library'
            typename TypeConv<D>::func_type   __cnv;

            // the original object
            PyObject*                         __ptr;
            // and the element size so the converters
            // will now how many bytes to advance to
            // go to the nxt element - this allows
            // for a more flexible arch
            unsigned int                      __esize;

            // no default c'tor, we really must have
            // an object
            conv();
            // and no copy/assignment either
            conv( const conv<D, AT>& );
            const conv<D, AT>& operator=(const conv<D, AT>&);
    };
}

// These defines + the use of the TYPELISTn() defines 
// will help setting up allowable conversion chains for
// POD and Complex datatypes

// a is a TypeList [Use TYPELISTn() macros for readable code ;)]
// b is the destination type of the cast
// c is the 'caster' to be used for casting types in the
//   typelist a to the destination type b, eg podcast, boolcast,
//   cmplxcast, stringcast, ..., see above.
// d is string which will be used to construct a variable name.
//   Make sure it is unique for each ENABLECASTFROM_TO_EXT()
//   use
#define ENABLECASTFROM_TO_EXT(a, b, c, d) \
    namespace arraysupport {\
        template <> \
        TypeConv<b>::map_type TypeConv<b>::__converters = TypeConv<b>::map_type();\
        Chain<a,b,c> cht##d##c = Chain<a,b,c>();\
    }
// shorthand defines for various typelists

// list of integral types
#define ILIST TYPELIST6(char, unsigned char, int,unsigned int, short, unsigned short)

// list of all supported/recognized POD types
#define PODLIST TYPELIST10(char, unsigned char, short, unsigned short, int, unsigned int, long, unsigned long, float, double)

// list of supported complex types
#define MCPLX TYPELIST5(support::mcplx<char>,support::mcplx<int>,support::mcplx<long>,support::mcplx<float>,support::mcplx<double>)

// integral types should be castable to bool + char
ENABLECASTFROM_TO_EXT(ILIST,bool, boolcast,a);

ENABLECASTFROM_TO_EXT(ILIST,char, podcast,b);
ENABLECASTFROM_TO_EXT(ILIST,unsigned char, podcast,c);

// plain-old-datatypes should be convertible
ENABLECASTFROM_TO_EXT(PODLIST, int, podcast,d);
ENABLECASTFROM_TO_EXT(PODLIST, unsigned int, podcast,e);
ENABLECASTFROM_TO_EXT(PODLIST, long, podcast,f);
ENABLECASTFROM_TO_EXT(PODLIST, unsigned long, podcast,g);
ENABLECASTFROM_TO_EXT(PODLIST, float, podcast,h);
ENABLECASTFROM_TO_EXT(PODLIST, double, podcast,i);

// conversions of complex numbers
ENABLECASTFROM_TO_EXT(MCPLX, casa::Complex, cmplxcast, j);
ENABLECASTFROM_TO_EXT(MCPLX, casa::DComplex, cmplxcast, k);

// and for string
ENABLECASTFROM_TO_EXT(TYPELIST3(char,unsigned char, signed char), casa::String, stringcast,l);

// not all num* implementations have the same signature for
// 'import_array()' (for loading the API)
// Numeric,numarray =>   void import_array()
// Numpy (scipy?)   =>   int  import_array()
// So this wrapper will just do stuff in such a
// way we will have
//   bool   pycasa_import_array()
bool pycasa_import_array() {
    bool    rv( (PyArray_API!=0) );

    if( !rv ) {
        import_array();
        rv = (PyArray_API!=0);
    }
    return rv;
}

// require loading of the numarray interface
// If that fails - bugger all!

#ifndef NUMTYPE
#define NUMTYPE "Unknown"
#endif

struct loadinit {
    // in the c'tor we do whatever we
    // want to do to initialize this thang
    loadinit() {
        cout << "loadinit: loading " << NUMTYPE << " array API" << endl;
        if( !pycasa_import_array() )
            throw failed_to_load_array_API();
    }

    // if you want to clean up stuff
    // put it in the d'tor
    // At the moment no cleaning req'ed
    // so no d'tor
};

// this triggers global initialization
static loadinit   do_loadinit = loadinit();



// mapping of casa::datatype -> python Numeric datatype
typedef std::map<casa::DataType,PyArray_TYPES>   dtmap_t;

dtmap_t make_dtmap( void ) {
    dtmap_t  rv;

    rv[casa::TpBool]       = PyArray_INT;
    rv[casa::TpInt]        = PyArray_INT;
    rv[casa::TpUInt]       = PyArray_UINT;
    rv[casa::TpString]     = PyArray_OBJECT;
    rv[casa::TpDouble]     = PyArray_DOUBLE;
    rv[casa::TpFloat]      = PyArray_FLOAT;
    rv[casa::TpComplex]    = PyArray_CFLOAT;
    rv[casa::TpDComplex]   = PyArray_CDOUBLE;
    rv[casa::TpArrayFloat] = PyArray_OBJECT;
    return rv;
}

dtmap_t   dtmap = make_dtmap();


// sjeezus... operator== on casa::IPosition
// throws!!! who-the-fucking-hell EVER thought
// that one up?!
// Write our own derived class which will
// redefine the comparison operators to
// call the non-throwing one.
class JPosition:
    public casa::IPosition
{
    public:
        JPosition():
            casa::IPosition()
        {}
        JPosition(casa::uInt l):
            casa::IPosition(l)
        {}
        JPosition(casa::uInt l, casa::Int v):
            casa::IPosition(l, v)
        {}
        
        // jesus-holy-tapdancing-Christ!
        // Even bloody-operator=() will throw?!?!
        const JPosition& operator=( const JPosition& other ) {
            this->resize(0);
            this->casa::IPosition::operator=(other);
            return *this;
        }
        const JPosition& operator=( const casa::IPosition& other ) {
            this->resize(0);
            this->casa::IPosition::operator=(other);
            return *this;
        }
};

// the versions of operator==
bool operator==(const JPosition& l, const JPosition& r) {
    return l.casa::IPosition::isEqual(r);
}
bool operator==(const JPosition& l, const casa::IPosition& r) {
    return l.casa::IPosition::isEqual(r);
}
bool operator==(const casa::IPosition& l, const JPosition& r) {
    return l.isEqual(r);
}
// id. for operator !=
bool operator!=(const JPosition& l, const JPosition& r) {
    return !(l==r);
}
bool operator!=(const JPosition& l, const casa::IPosition& r) {
    return !(l==r);
}
bool operator!=(const casa::IPosition& l, const JPosition& r) {
    return !(l==r);
}






// =======================================
//
//     Start support for sequences here
//
// =======================================
namespace sequencesupport {
    static JPosition   noshape = JPosition();

    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    //
    //     start with abstracting sequence access
    //     plus wrapping (possible) #defines.
    //     #defines won't fit in as template
    //     parameters whilst functions do..
    //
    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    
    // same as in support::py*_check to
    // translate #defines into 'real'
    // functions but now for sequence types
    int pynumarray_check(PyObject* o) {
        return PyArray_Check(o);
    }
    // return true if it's a numarray of objects
    int pynumarray_object_check(PyObject* o) {
        if( pynumarray_check(o) ) {
            return pyarrayisobject(o)?(1):(0);
        }
        return 0;
    }
    int pytuple_check(PyObject* o) {
        return PyTuple_Check(o);
    }
    int pylist_check(PyObject* o) {
        return PyList_Check(o);
    }

    // size functions
    int pylist_size(PyObject* o) {
        return PyList_Size(o);
    }
    int pytuple_size(PyObject* o ) {
        return PyTuple_Size(o);
    }
    // from an nd-array of objects, get the
    // size of the first dim
    int pynumarray_object_size(PyObject* o) {
        return ((pyarrayndim(o)>0)?(pyarraydim(o, 0)):(0));
    }

    // element access
    PyObject* pylist_getitem(PyObject* o, int idx) {
        return PyList_GetItem(o, idx);
    }
    PyObject* pytuple_getitem(PyObject* o, int idx) {
        return PyTuple_GetItem(o, idx);
    }
    // should only be used on numarrays of objects
    PyObject* pynumarray_object_getitem(PyObject* o, int idx) {
        return *(PyObject**)pyarraygetptr1(o, idx);
    }

    // Abstract out the details of accessing elements into
    // a sequence policy.
    // The python type must support the following thingys:
    //   int checkfn(PyObject*)
    //        should return 0 if not correct type
    //   int sizefn(PyObject*)
    //        return the 1D size of the thingy
    //   PyObject* getfn(PyObject*,int)
    //        return the n'th item out of the sequence
    template <int (*checkfn)(PyObject*),
              int (*sizefn)(PyObject*),
              PyObject* (*getfn)(PyObject*, int)>
    struct sequence_policy {
        static int typecheck(PyObject* o) {
            return checkfn(o);
        }
        static int size(PyObject* o) {
            return sizefn(o);
        }
        static PyObject* getitem(PyObject* o, int n) {
            return getfn(o, n);
        }
    };
    typedef sequence_policy<pylist_check, pylist_size, pylist_getitem>    list_policy;
    typedef sequence_policy<pytuple_check, pytuple_size, pytuple_getitem> tuple_policy;
    typedef sequence_policy<pynumarray_object_check,
                            pynumarray_object_size,
                            pynumarray_object_getitem>                    objarray_policy;

    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    //            
    //       Step one: retrieve the shape from a PyObject.
    //       Want to support nested sequences, eg
    //       l1 = [1,2,3]
    //       t1 = (4,5,6)
    //       c  = array([l1, t1])
    //
    //       shape(c) should return (2,3) 
    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

    // we must be able to get the shape of a sequence
    // the shaper interface must take a PyObject* and
    // get a JPosition from that.
    // Note: we declare the function prototypes here
    //       and we stick them in a template
    //       but... we will fill the functions in
    //       later on! Neat eh?
    JPosition numarray_shape( PyObject* o );
    JPosition tuple_shape( PyObject* o );
    JPosition list_shape( PyObject* o );
    // if the object is a sequence of scalars
    // that's acceptable as well...
    JPosition scalar_shape( PyObject* o );

    // this is the template shaper which will
    // try all alternative shapers in turn
    // until either exhausted or a successfull
    // shape was returned
    typedef JPosition (*sfnptr)(PyObject*);
    template <sfnptr S1, sfnptr S2, sfnptr S3, sfnptr S4>
    struct shaper {
        static JPosition get_shape(PyObject* o) {
            JPosition              r;
            if( o==0 )
                return noshape;

            // check all alternatives
            if( (r=S1(o))==noshape ) {
                if( (r=S2(o))==noshape ) {
                    if( (r=S3(o))==noshape ) {
                        r=S4(o);
                    }
                }
            }
            return r;
        }
    };
    // here we typedef the previous template into
    // a Shaper.
    // Now we can start using Shaper::get_shape(PyObject*)
    // even though the actual functions haven't been
    // defined yet :) cool eh?
    // No really - now we can easily recurse: in the
    // function definitions we can refer to this
    // typedef...!
    typedef shaper<scalar_shape, numarray_shape, tuple_shape, list_shape> Shaper;
  
    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    //
    //            the real shapers start here
    // 
    // @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


    // A generic method for getting the shape
    // of an object which can be abstracted
    // by a sequence_policy (SP)
    //   Note: will recursively check all dims...
    //         ie supports heterogenous combinations
    //         of sequence thingies
    template <typename SP>
    struct pyobj_shape {
        // the static fn
        static JPosition get_shape( PyObject* o ) {
            // if not the type we need
            if( !SP::typecheck(o) )
                return noshape;
            // ok do stuff
            JPosition rv;
            JPosition myshape;

            // we're supposed to be a 1D sequence
            // with some size
            myshape = JPosition(1, SP::size(o));
        
            // check if we need to check anything at all
            if( myshape[0]==0 )
                return myshape;

            // ok there are elements to check. Assert
            // that all elements have the same shape.
            int          idx;
            PyObject*    ob0;
            JPosition    shp0;

            ob0  = SP::getitem(o, 0);
            Py_XINCREF(ob0);
            shp0 = Shaper::get_shape(ob0);
            Py_XDECREF(ob0);
            for( idx=1; idx<myshape[0]; idx++ ) {
                PyObject*  optr;
                JPosition  shpn;

                optr = SP::getitem(o, idx);
                Py_XINCREF(optr);
                shpn = Shaper::get_shape(optr);
                Py_XDECREF(optr);
                if( shpn!=shp0 )
                    break;
            }
            if( idx!=myshape[0] ) {
                return noshape;
            }
            // Now construct the final return-value:
            // we are a shp0.ndim()+1 array
            // with dim[0]==myshape[0] and
            // the other dims are shp0
            myshape.append( shp0 );
            return myshape;
        }
    };

    // the shape of a numarray
    JPosition numarray_shape( PyObject* o ) {
        // if it's a numarray of objects, we may shortcut here
        if( pynumarray_object_check(o) )
            return pyobj_shape<objarray_policy>::get_shape(o);
            
        // check if the object is a numarray thing at all
        if( !pynumarray_check(o) )
            return noshape;

        int             nd;
        JPosition       rv;

        // check if there's anything to check -
        // if ndim==0 there's nothing to do, really
        nd = pyarrayndim(o);

        if( nd<=0 )
            return noshape;

        // array of scalars
        // can just copy over the dims
        int             pyidx;
        unsigned int    casaidx;

        rv = JPosition( nd );
        for( pyidx=0, casaidx=0; pyidx<nd; ++pyidx, ++casaidx ) {
            rv[casaidx] = pyarraydim(o, pyidx);
        }
        return rv;
    }

    // the functions basically use the generic thing
    // from above
    JPosition tuple_shape(PyObject* o) {
        return pyobj_shape<tuple_policy>::get_shape(o);
    }
    JPosition list_shape(PyObject* o) {
        return pyobj_shape<list_policy>::get_shape(o);
    }
    
    // for the scalar shaper, need to do a bit more work:

    // define what we accept as scalartypes
    typedef int (*cfn)(PyObject*);
    template <cfn S1, cfn S2, cfn S3, cfn S4, cfn S5, cfn S6, cfn S7>
    struct is_scalar {
        static int do_test(PyObject* o) {
            int  r;
            r = S1(o);
            (void)(r==0 && (r=S2(o)));
            (void)(r==0 && (r=S3(o)));
            (void)(r==0 && (r=S4(o)));
            (void)(r==0 && (r=S5(o)));
            (void)(r==0 && (r=S6(o)));
            (void)(r==0 && (r=S7(o)));
            return r;
        }
    };
    typedef is_scalar<support::pyint_check,
                      support::pylong_check,
                      support::pystring_check,
                      support::pyfloat_check,
                      support::pybool_check,
                      support::pycomplex_check,
                      support::pydict_check> Scalar;

    // Function template:
    // the scalar-shaper uses
    // the previous asserter, and asserts that
    //  1) the pyobject is a certain (sequence) type
    //  2) all of its elements are scalars
    // Uses the SequencePolicy stuff from above
    template <typename SP>
    JPosition scalar_asserter(PyObject* o) {
        if( !SP::typecheck(o) )
            return noshape;
        JPosition  myshape( 1, SP::size(o) );
        
        if( myshape[0]==0 )
            return myshape;
        int   idx; 
        for(idx=0; idx<myshape[0]; idx++) {
            int        sclr;
            PyObject*  optr;
            optr = SP::getitem(o, idx);
            Py_XINCREF(optr);
            sclr = Scalar::do_test(optr);
            Py_XDECREF(optr);
            if( !sclr )
               break;
        }
        if( idx!=myshape[0] )
            myshape = noshape;
        return myshape;
    }
    
    JPosition scalar_shape(PyObject* o) {
        JPosition              r;
        if( (r=scalar_asserter<tuple_policy>(o))==noshape ) {
             r=scalar_asserter<list_policy>(o);
        }
        return r;
    }

    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    //
    //  This is going to be interesting...
    //  
    //  Given a PyObject*, which may consist of nested
    //  sequences/numarrays and containing heterogenous
    //  objects (except for numarrays-with-POD-in-them),
    //  (attempt to) transform this bunch-o-stuff
    //  into a nice casa::Array<some-desired-type> ...
    //
    //  We're going to assume that the Shape-stuff will
    //  test for the 'layout' of the array, that is, the
    //  whole bunch-o-stuff has a well-defined shape.
    //  Eg:
    //    given:
    //      a = [1,2]; b=(3,4,5); c=[a,b]
    //    then our shape-stuff should NOT return
    //    a shape (it returns sequencesupport::noshape) since
    //    len(c[0]) != len(c[1])
    //    whilst it is perfectly valid python.
    //
    // We must also be prepared to accept varying
    // python-types: PyList/PyTuple are heterogenous
    // and not every element has to have the correct
    // type. No problem as long as we can transform
    // the contained python object into something 
    // we want.
    //
    // Then there is the fact that plain-old-C datatypes
    // do not necessarily map 1-to-1 to casa datatypes...
    // and the puzzle is even more nice.
    //
    // Solution:
    //
    //  Give user a chance to specify which casatype
    //  he/she wants to end up with and give him/her
    //  the possibility to define an intermediate type
    //  for which both the conversion from Python as
    //  well the conversion to casa is well defined.
    //
    // In short, we do the following conversion chain:
    //
    // PyObject* -> intermediate, well-defined type ->
    //     potentially different casa datatype
    //
    //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

    // a nofill policy. don't do nothing and always return false
    template <typename>
    struct nofill {
        template <typename CT>
        static bool do_fill(casa::Array<CT>&, PyObject*, const JPosition&) {
            return false;
        }
    };
   
    // reserve for five different fillers - can be easily expanded
    // later on
    template <template <typename> class D1 = nofill,
              template <typename> class D2 = nofill,
              template <typename> class D3 = nofill,
              template <typename> class D4 = nofill,
              template <typename> class D5 = nofill>
    struct basefiller {
        template <typename AT>
        struct astype {
            template <typename CT>
            static bool do_fill(casa::Array<CT>& arr,
                                PyObject* o,
                                const JPosition& start = JPosition()) {
                bool                     retval;
                typename AT::return_type  v;
                // if we can sucessfully convert the current thingy to
                // the requested scalar type, this must be the
                // end of the recursion; set the value at the
                // current arrayposition to what it's supposed to be
                //
                // Before we try to do the conversion, we must
                // assert that it is a scalar
                if( (Scalar::do_test(o)!=0 &&
                    ((retval=AT::do_convert(v, o))==true))==true ) {
                    // set the value at position start to the
                    // converted value
                    // first convert from intermediate type
                    // (AT::return_type) to the requested casatype (CT)
                    CT   cv( v );
                    arr( start ) = cv;
                } else {
                    // try the alternatives
                    retval = D1<AT>::do_fill(arr, o, start);
                    (void)(!retval && (retval=D2<AT>::do_fill(arr, o, start)));
                    (void)(!retval && (retval=D3<AT>::do_fill(arr, o, start)));
                    (void)(!retval && (retval=D4<AT>::do_fill(arr, o, start)));
                    (void)(!retval && (retval=D5<AT>::do_fill(arr, o, start)));
                }
                return retval;
            }
        };
    };
    template <typename T> struct pylist_filler;
    template <typename T> struct pytuple_filler;
    template <typename T> struct pynumarray_object_filler;
    template <typename T> struct pynumarray_filler;

    typedef basefiller<pylist_filler,
                       pytuple_filler,
                       pynumarray_object_filler,
                       pynumarray_filler>  Basefiller;


    // template for a function which
    // processes a sequence_policy (SP)
    // for a given intermediate type (AT)
    //  (look in the warp:: namespace, the as_<something>
    //   thingies)
    // and for an Array of some type (CT)
    //  [CT will be deduced from argument deduction ;)
    //   don't have to specify]
    template <typename AT, typename SP>
    struct seqfiller {
        template <typename CT>
        static bool do_fill(casa::Array<CT> arr, PyObject* o, const JPosition& pos) {
            if( !SP::typecheck(o) )
                return false;
            int          i;
            int          nel( SP::size(o) );
            JPosition    cur( pos );
            // IPosition uses 0-based indexing so the size before
            // we add 1 to the number of dimensions is the index
            // of the axis WE will loop over
            unsigned int myidx( pos.nelements() );

            // ok, append our dim to the current position
            cur.append( casa::IPosition(1) );

            // and process the elements
            for(i=0; i<nel; ++i) {
                PyObject*   optr;

                optr       = SP::getitem(o, i);
                cur[myidx] = i;
                // if failure to fill - get the beep out
                if( !Basefiller::astype<AT>::do_fill(arr, optr, cur) )
                    break;
            }
            return (i==nel);
        }
    };

    // now we can use the template to actually define
    // the fillers for concrete types
    template <typename AT>
    struct pylist_filler {
        template <typename CT>
        static bool do_fill(casa::Array<CT>& arr, PyObject* o, const JPosition& pos) {
            // yep use the template
            return seqfiller<AT, list_policy>::do_fill(arr, o, pos);
        }
    };
    template <typename AT>
    struct pytuple_filler {
        template <typename CT>
        static bool do_fill(casa::Array<CT>& arr, PyObject* o, const JPosition& pos) {
            // yep use the template
            return seqfiller<AT, tuple_policy>::do_fill(arr, o, pos);
        }
    };
    // while we're at it, define a filler for
    // a numarray of objects... (eg strings!)
    template <typename AT>
    struct pynumarray_object_filler {
        template <typename CT>
        static bool do_fill(casa::Array<CT>& arr, PyObject* o, const JPosition& pos) {
            // yep use the template
            return seqfiller<AT, objarray_policy>::do_fill(arr, o, pos);
        }
    };
   
    // finally, an (optimized) filler for a generic numarray object
    // containing POD
    // Note: the AT is not used here - numarrays can only
    // contain POD, right?
    template <typename AT>
    struct pynumarray_filler {
        template <typename CT>
        static bool do_fill(casa::Array<CT>& arr, PyObject* o, const JPosition& pos ) {
            if( !pynumarray_check(o) )
                return false;
            // ok, so it is a numarray we're stuffing into
            // some casa::Array...
            JPosition         shp( numarray_shape(o) );

            if( shp==noshape || shp.product()==0 )
                return false;

            // alrighty then, we do not completely discard
            // the intermediate type; we take the type from AT
            // but do not use that warper - we use a different one
            typedef arraysupport::podcast<CT, typename AT::return_type> IType;
            JPosition                     b, e, m;
            countedarray<CT>              converteddata;
            arraysupport::conv<CT, IType> converter( o );

            // hmmm casa::Array supports slice-assignment!
            // So if we could manage to create a casa::Array
            // from the arraydata in the numarray - we could
            // copy the whole lot in one go!
            
            // Firstattempt the conversion
            converteddata = converter();
            if( ((CT*)converteddata)==0 ) {
                cerr << "pynumarray_filler: failed to convert from PyArray_TYPE "
                     << "to requested casa-type" << endl;
                return false;
            }

            // the IPositions
            //   we have the start IPosition (pos)
            // * add a JPosition of ourdim 0's to it;
            //   that's the start IPos.
            // * for end we append just our shape to pos
            b = pos.concatenate( casa::IPosition(shp.nelements(), 0) );
            e = pos.concatenate( shp-1 );
            // m == myshape, prefixed with pos.nelements() degenerate
            //      axes -> otherwise the arrayassignment fails because
            //      it fails to recognize that the total number of
            //      elements are equal... Stoopid b*sterds!
            m = casa::IPosition( pos.nelements(), 1 ).concatenate( shp );

            // and now [drum roll...] the MAGIC part ;)
            arr(b, e) = casa::Array<CT>(m, (CT*)converteddata, casa::SHARE);

            // done!
            return true;
        }
    };
}

// stick some conversion (warp) functors
// into their own namepsace.
// These functors can be used to transform
// something into a compile-time known returntype.
// These 'warpers' are used to transform
// the elements of casa arrays into their
// numarray supported equivalent layout.
// Typically, Python numarray objects are
// plain-old-C-style arrays and for some
// casa datatypes we MUST make sure that
// the layout in the numarray object adheres
// to the layout numarray expects.
namespace warp {
    template <typename T>
    struct identity {
        typedef T  returntype;
        T operator()( const T& t ) const {
            return t;
        }

        // warp a number of objects in one go
        // Since this is the *identity* warper
        // it assumes T is POD [plain old data]
        // hence that bitwise copies are Ok
        // For objects which must use 'operator='
        // to work properly, we must implement
        // a different warper. For now it's not
        // necessary so...
        void operator()(T* dst, const T* src, unsigned int n) const {
            ::memcpy(dst, src, n*sizeof(T));
            return;
        }
    };

    // casa::Bool is represented as int  (PyArray_INT)
    // so we warp a casa::Bool into that. The identity
    // relation (warp::identity) didn't apparently hold
    // very well for this type...
    template <typename T>
    struct casabool2int {
        typedef int   returntype;
        
        // only support warping of casa::Bool for now...
        int operator()( const casa::Bool& b ) const {
            return (b==casa::True)?(1):(0);
        }

        // a number of bools in one go
        void operator()(int* dst, const casa::Bool* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
        }
    };

    // the template parameter here is unused but at least
    // now this thang can be used as template template
    // parameter
    template <typename T=int>
    struct ToPyStr {
        typedef PyObject* returntype;
        PyObject* operator()( const casa::String& t ) const {
            return PyString_FromString(t.chars());
        }

        // nothing else to do but to loop over all items...
        void operator()(returntype* dst, const casa::String* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
            return;
        }
    };

    // for casa::Complex and casa::DComplex

    // this warper ignores the template argument
    // and converts whatever is passed to it
    // to an 'array of 2 floats'
    template <typename T>
    struct ToPyComplex {
        typedef float basic_type;
        typedef struct {
            basic_type   re;
            basic_type   im;
        } returntype;

        // Support anything that has .real() and .imag()
        // which are convertible to 'float'
        template <typename U>
        returntype operator()( const U& c ) const {
            returntype    rv;

            rv.re = basic_type(c.real());
            rv.im = basic_type(c.imag());
            return rv;
        }

        // first, the generic, slow implementation
        // which supports converting an array of
        // U's which have .real() and .imag()
        // which are convertible to float.
        template <typename U>
        void operator()(returntype* dst, const U* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
            return;
        }

        // specialization for casa::Complex -> these are
        // in layout identical to PyComplex so we can
        // optimize now and just do a plain memcpy!
        void operator()(returntype* dst, const casa::Complex* src, unsigned int n) const {
            ::memcpy(dst, src, n*sizeof(returntype));
            return;
        }
    };

    template <typename T>
    struct ToPyDComplex {
        typedef double basic_type;
        typedef struct {
            basic_type   re;
            basic_type   im;
        } returntype;

        // Support anything that has .real() and .imag()
        // which are convertible to 'double'
        template <typename U>
        returntype operator()( const U& c ) const {
            returntype    rv;

            rv.re = basic_type(c.real());
            rv.im = basic_type(c.imag());
            return rv;
        }

        // first, the generic, slow implementation
        // which supports converting an array of
        // U's which have .real() and .imag()
        // which are convertible to double.
        template <typename U>
        void operator()(returntype* dst, const U* src, unsigned int n) const {
            for( unsigned int i=0; i<n; ++i ) {
                *dst++ = (*this)(*src++);
            }
            return;
        }

        // specialization for casa::DComplex -> these are
        // in layout identical to PyDComplex so we can
        // optimize now and just do a plain memcpy!
        void operator()(returntype* dst, const casa::DComplex* src, unsigned int n) const {
            ::memcpy(dst, src, n*sizeof(returntype));
            return;
        }
    };


    // section of warpers which take a PyObject
    // and transform into POD datatype 'T'.

    // define functor template for fn which
    // checks/extracts a basic type from
    // a python object (exposes the 'do_convert(T&, PyObject*)'
    // interface
    //
    // This should only go for
    //   long + PyInt_AsLong
    //   long + PyLong_AsLong
    //   double + PyFloat_AsDouble
    //   bool + support::pybool_ascppbool
    //   Py_complex + PyComplex_AsCComplex
    //   char* + PyString_AsString
    template <typename T, int (*checkfn)(PyObject*), T (*convfn)(PyObject*)>
    struct fromPythonBasic {
        typedef T return_type;
        static bool do_convert(T& t, PyObject* o ) {
            bool rv( false );
            if( checkfn(o) ) {
               t  = T(convfn(o));
               rv = true;
            }
            return rv;
        }
    };

    // this one accepts all calls and always
    // returns false. Useful for non-existant
    // alternatives. We must be prepared to 
    // be able to accept alternatives but 
    // sometimes there just aren't any.
    // For those cases, use this'un
    template <typename T=int>
    struct no_alternative {
        typedef T return_type;

        template <typename U>
        static bool do_convert( U& , PyObject* ) {
            return false;
        }
    };

    // this is a template which allows you
    // to specify alternate conversions
    // if you go from Python->plain-old-c++
    template <typename T,
              typename C1 = no_alternative<T>,
              typename C2 = no_alternative<T>,
              typename C3 = no_alternative<T> >
    struct fromPython {
        typedef T  return_type;

        // also exposes the 'do_convert(T&, PyObject*)'
        // method but tries the alternatives in sequence
        // until succesfull or alternatives exhausted
        static bool do_convert(T& t, PyObject* o ) {
            bool                     rv;
            typename C1::return_type c1;

            if( (rv=C1::do_convert(c1, o))==true ) {
                t  = T(c1);
            } else {
                typename C2::return_type  c2;
                if( (rv=C2::do_convert(c2, o))==true ) {
                    t = T(c2);
                } else {
                    typename C3::return_type  c3;
                    if( (rv=C3::do_convert(c3, o))==true ) {
                        t = T(c3);
                    }
                }
            }
            return rv;
        }
    };
    // short-hand typedefs for basic conversions
    // read something as "_ial" to mean "PyInt-as-long"
    // or "_laul" to mean "PyLong-as-unsigned-long"
    typedef fromPythonBasic<long, support::pyint_check, PyInt_AsLong> fpb_ial;
    typedef fromPythonBasic<long, support::pylong_check, PyLong_AsLong> fpb_lal;
    typedef fromPythonBasic<unsigned long, support::pyint_check, PyInt_AsUnsignedLongMask> fpb_iaul;
    typedef fromPythonBasic<unsigned long, support::pylong_check, PyLong_AsUnsignedLong> fpb_laul;
    typedef fromPythonBasic<double, support::pyfloat_check, PyFloat_AsDouble> fpb_dad;
    typedef fromPythonBasic<bool, support::pybool_check, support::pybool_ascppbool> fpb_bab;
    typedef fromPythonBasic<Py_complex, support::pycomplex_check, PyComplex_AsCComplex> fpb_cac;
    typedef fromPythonBasic<char*, support::pystring_check,PyString_AsString>  fpb_sas;

    // objects who have __str__ may also be convertible to string ;)
    // Make use of the already defined 'frompythonbasic String as string'
    // thingy
    int    has_str_method(PyObject* o) {
        return PyCallable_Check( PyObject_GetAttrString(o, "__str__") );
    }
    // can do this safely since the checkfn has
    // already verified that this can be done
    char*  call_str_method( PyObject* o ) {
        char*       rv=0;
        PyObject*   strrep = PyObject_CallMethod(o, "__str__", 0);

        fpb_sas::do_convert(rv, strrep);
        return rv;
    }
    typedef fromPythonBasic<char*, has_str_method, call_str_method>           fpb_callstr;


    // Now, from the basic conversions, we may build conversion chains/alternatives.
    // We may accept different source python types for the same destination
    // C/C++ type
    // Eg: for python -> int we support both PyInt & PyLong
    typedef fromPython<int,          fpb_ial, fpb_lal>          as_int;
    typedef fromPython<unsigned int, fpb_iaul, fpb_laul>        as_uint;
    // for float/double we accept PyFloat but also PyInt and PyLong
    typedef fromPython<float,        fpb_dad, fpb_ial, fpb_lal> as_float;
    typedef fromPython<double,       fpb_dad, fpb_ial, fpb_lal> as_double;

    // for bool we support PyBool, PyInt and PyLong
    typedef fromPython<bool,        fpb_bab, fpb_ial, fpb_lal>  as_bool;

    // string
    typedef fromPython<std::string, fpb_sas>                    as_stringexact;
    typedef fromPython<std::string, fpb_sas, fpb_callstr>       as_string;

    // For complex we go from PyObject*[PyComplex] -> Py_complex (C-struct with 2 doubles)
    // to intermediate type 'mcplx<T>' which allows converting from
    // Py_complex to any other complex type (int? float? double?)
    // The mcplx<T> has '.real()' and '.imag()' methods for extracting.. well, real+imag.
    // The mcplx<T> has a constructor from Py_complex which (attempts) the
    //  conversion from double to T.
    typedef fromPython<support::mcplx<float>, fpb_cac>          as_fcomplex;
    typedef fromPython<support::mcplx<double>, fpb_cac>         as_dcomplex;  
}


//
// Transform casa::Array to a numarray
// 
template <template <typename> class Transform=warp::identity>
struct arraytransformer {
    template <typename T>
    PyObject* operator()( const casa::Array<T>& t ) const
    {
        Transform<T>                        warper;
        PyArrayObject*                      po;
        casa::IPosition                     shp( t.shape() );
        casa::IPosition                     newshp( shp.nelements() );
        dtmap_t::const_iterator             dstdt;
        typename Transform<T>::returntype*  dst;

        // if we don't recognize the type, tell 'm so!
        // The datatypemap tells us, given the casa datatype,
        // what the corresponding Numarray datatype would be
        if( (dstdt=dtmap.find(casa::whatType((T*)0)))==dtmap.end() )
        {
            // if this occurs, maybe you need to 
            // add an entry to the 'dtmap' variable....
            // browse this file and look for it.
            ostringstream s;
            s << __FILE__ << ":" << __LINE__ 
              << " - arraytransformer<> -> Unhandled datatype "
              << casa::whatType((T*)0);
            cerr << s.str() << endl;
            PyErr_SetString(PyExc_ValueError,s.str().c_str());
            throw boost::python::error_already_set();
        }

        // ok, so we recognized the type, only the array's empty...
        if( shp.product()==0 ) {
            Py_INCREF(Py_None);
            return Py_None;
        }

        // create the array from the dimensions
        po = (PyArrayObject*)PyArray_FromDims(shp.nelements(),
                                              (int*)shp.storage(),
                                              dstdt->second);

        if( po==0 ) {
            PyErr_SetString(PyExc_MemoryError, "Failed to allocate Numarray-object");
            throw boost::python::error_already_set();
        }

        // do a cast to destination type
        dst = (typename Transform<T>::returntype*)po->data;

        // do it like this to be able to *transform*
        // between different types...
        // Eg when transforming an array of strings;
        // each string has to be transformed to
        // a PyObject in that case...
        //
        // So we first get a contiguous array
        // out of the source array such that
        // we can loop as fast as possible
        // over the elements.
        //
        // The 'warper' will be called upon
        // to do the transformation.
        // There may be specializations
        // which do a memcpy, others use 
        // a for-loop with operator=()
        // [sometimes you just have to do it like
        //  that...]
        const T*   src;
        casa::Bool del;

        src = t.getStorage(del);
        warper(dst, src, shp.product());
        t.freeStorage(src, del);

        // copy the strides to fix
        // array-indexing
        // numarray strides are in units of bytes,
        // casa steps are in units of 'items'
        // so must multiply the casa strides by
        // the itemsize
        for( unsigned int k=0; k<t.steps().nelements(); ++k )
            po->strides[k] = t.steps()[k] * sizeof(typename Transform<T>::returntype);

        return (PyObject*)po;
    }
};


// convert all the arrayvalues
struct tostdstring {
    std::string operator()( const casa::String& cs ) const {
        return std::string(cs.chars());
    }
};


// Record-field accessor, gets the (scalar) valued field
// into the variable 'value'
template <typename T>
struct rfaccessor_t {
    rfaccessor_t(const casa::RecordInterface& r, const casa::RecordFieldId& fi) {
        r.get(fi, value);
    }

    T   value;
};

// id. as above, now for arrayvalued fields
// into the variable 'value'
template <typename T>
struct rfaaccessor_t {
    rfaaccessor_t(const casa::RecordInterface& r, const casa::RecordFieldId& fi) {
        r.get(fi, value);
    }

    casa::Array<T>   value;
};

// function template for converting casa (scalar) valued RecordFields into
// a PyObject, using an intermediate conversion to allow translation from
// (potentially not supported by Python)casa types to
// python supported built-in types.
// E.g. the actual conversion chain of 
//      casa::Bool -> PyObject*
// goes like:
//      casa::Bool -> built-in type 'long' -> PyBool_FromLong -> PyObject*
template <typename T, typename U, PyObject* (*fptr)(U)>
struct gen {
    static PyObject* do_gen(const casa::RecordInterface& r, casa::Int i) {
        rfaccessor_t<T>  rfa(r, i);
        return fptr(U(rfa.value));
    }
};

// id. for array-valued thingies (hence the 'a' suffix)
// Specify a 'Transform' template template parameter;
// this class will be used to transform individual array
// elements to a corresponding python-numarray-type
template <typename T, template <typename> class Transform = warp::identity>
struct gena {
    static PyObject* do_gen(const casa::RecordInterface& r, casa::Int i) {
        rfaaccessor_t<T>  rfa(r, i);
        return arraytransformer<Transform>()(rfa.value);
    }
};

// function pointer which defines an interface which 
// takes a "RecordInterface" and a fieldnumber
// and warps it into a PyObject*
typedef PyObject* (*genfunc)(const casa::RecordInterface&, casa::Int num);

// map casa datatype to a pointer-to-function so the system
// knows how to transform a casa datatype into a PyObject
typedef map<casa::DataType,genfunc> genmap_t;


PyObject* make_pystr( const casa::String& s ) {
    return PyString_FromString(s.chars());
}

// because of 'circular' dependency, or, recursiveness if you will,
// the function which turns a (Table)Record into a PyDict must be able
// to find itself in the global 'genmap_t' variable, but from the
// function we must be able to reference the global genmap_t variable,
// decide to have a placeholder here which we can stick into the map.
PyObject* make_pydict_fromsubrecord(casa::RecordInterface const&, casa::Int num);

PyObject* make_pyobject_fromsubtable(casa::RecordInterface const&, casa::Int num);


// fill in the mapping
genmap_t make_genmap() {
    genmap_t   rv;

    rv[casa::TpBool] = &gen<casa::Bool,long,PyBool_FromLong>::do_gen;
    rv[casa::TpInt] =  &gen<casa::Int,long,PyInt_FromLong>::do_gen;
    rv[casa::TpUInt] = &gen<casa::uInt,long,PyInt_FromLong>::do_gen;
    rv[casa::TpFloat] = &gen<casa::Float,double,PyFloat_FromDouble>::do_gen;
    rv[casa::TpDouble] = &gen<casa::Double,double,PyFloat_FromDouble>::do_gen;
    rv[casa::TpString] = &gen<casa::String,const casa::String&,make_pystr>::do_gen;

    rv[casa::TpRecord] = &make_pydict_fromsubrecord;
    rv[casa::TpTable]  = &make_pyobject_fromsubtable;

    rv[casa::TpArrayBool] = &gena<casa::Bool>::do_gen;
    rv[casa::TpArrayInt]  = &gena<casa::Int>::do_gen;
    rv[casa::TpArrayFloat] = &gena<casa::Float>::do_gen;
    rv[casa::TpArrayDouble]= &gena<casa::Double>::do_gen;
    rv[casa::TpArrayComplex] = &gena<casa::Complex, warp::ToPyComplex>::do_gen;
    rv[casa::TpArrayDComplex] = &gena<casa::DComplex, warp::ToPyDComplex>::do_gen;
    rv[casa::TpArrayString]   = &gena<casa::String, warp::ToPyStr>::do_gen;

    return rv;
}
genmap_t   genmap = make_genmap();



namespace boost { namespace python {
    // convert  casa String to PyString
    template <>
    struct to_python_value<casa::String const&> :
        detail::builtin_to_python
    {
        inline PyObject* operator()(casa::String const& cs) const {
            return ::PyString_FromString(cs.chars());
        }
    };

    // convert casa::Record to PyDict
    template <>
    struct to_python_value<casa::RecordInterface const&>:
        detail::builtin_to_python
    {
        inline PyObject* operator()( casa::RecordInterface const& ri) const {
            PyDictObject*   dict = (PyDictObject*)PyDict_New();

            if( !dict ) {
                return NULL;
            }

            // loop over all recordfields and transform them into
            // appropriate PythonObject* 's
            for( casa::uInt fld=0; fld<ri.nfields(); fld++ ) {
                genmap_t::iterator curgen( genmap.find(ri.type((casa::Int)fld)) );

                if( curgen!=genmap.end() ) {
                    PyDict_SetItem(  (PyObject*)dict,
                                     warp::ToPyStr<>()( ri.name((casa::Int)fld) ),
                                     curgen->second( ri, (casa::Int)fld ) );
                } else {
                    cout << "Unknown fieldtype for " << ri.name((casa::Int)fld)
                         << " (type=" << ri.type((casa::Int)fld) << ")" << endl;
                }
            }
            return (PyObject*)dict;
        }
    };

}} // boost::python



// And do it the other way around ->
// Convert from python string to
// casa::String
struct CasaStringFromPythonString {
    CasaStringFromPythonString() {
        converter::registry::push_back(&convertible, &construct, type_id<casa::String>() );
    }
    static void* convertible( PyObject* ptr ) {
        if( PyString_Check(ptr)==0 ) 
            return 0;
        return ptr;
    }
    static void construct( ::PyObject* pyptr,
            boost::python::converter::rvalue_from_python_stage1_data* data ) {
        const char* value = ::PyString_AsString(pyptr);

        if( value==0 )
            boost::python::throw_error_already_set();
        void* storage =
            ((boost::python::converter::rvalue_from_python_storage<casa::String>*)data)->storage.bytes;
        new (storage) casa::String(value);
        data->convertible = storage;
    }
};


// access a value in row 'r' in a scalar column 'c'
// in table 't'.
// Put the result into 'value'
template <typename T>
struct scolaccessor {
    scolaccessor(const casa::Table& t, const std::string& c, casa::uInt r) {
        casa::ROScalarColumn<T>(t, c).get(r, value);
    }

    T            value;
};

// Give access to a (scalar)column such that we can update a value
// at row 'r'. This may sound weird but at least you can't
// influence the value in a different row after you've created
// this baby
template <typename T>
struct scolputter {
    scolputter(casa::Table& tab, const std::string& col, casa::uInt row, const T& val ) {
        casa::ScalarColumn<T>    scol(tab, col);
        scol.put(row, val);
    }
};

// id. but now for an ArrayColumn
template <typename T>
struct acolaccessor {
    acolaccessor(const casa::Table& t, const std::string& c, casa::uInt r) {
        casa::ROArrayColumn<T>(t, c).get(r, value);
    }

    casa::Array<T>       value;
};
// and arrayvalued column put()
template <typename T>
struct acolputter {
    acolputter(casa::Table& tab, const std::string& col, casa::uInt row, const casa::Array<T>& val ) {
        casa::ArrayColumn<T>    scol(tab, col);
        scol.put(row, val);
    }
};

// Take all the values in column 'c' in table 't'
// and put the result in 'values'
// Defaults to ROScalarColumn + Vector; reading
// all the scalar values in a column into a vector.
// Can easily be used to get all the values in an
// array-valued column by changing the template
// parameters to
// ROArrayColumn + Array ....
// (big hoooorraaaaay up c++'s ass!)
template <typename T,
          template <typename> class colType = casa::ROScalarColumn,
          template <typename> class arrType = casa::Vector >
struct colaccessor2 {
    colaccessor2(const casa::Table& t, const std::string& c) {
        colType<T>(t, c).getColumn(values, casa::True);
    }
    arrType<T>   values;
};

// write the whole column in one go
// can use ArrayColumn<T> + Array<T> to
// set arrayvalued columns...
template <typename T,
          template <typename> class colType = casa::ScalarColumn,
          template <typename> class arrType = casa::Vector >
struct colputter2 {
    colputter2(const casa::Table& t, const std::string& c, const arrType<T>& values) {
        colType<T>(t, c).putColumn(values);
    }
};



// Tranform a PyObject into a POD datatype
// T is one of the as_<desiredtyped> 'policies'
// defined above
template <typename T>
struct ScalarFromPython {
    ScalarFromPython(PyObject* o) {
       if( T::do_convert(value, o)==false ) {
           throw incompatible_pythontype_for_columndatatype();
       }
    }
    typename T::return_type  value;
};

// id. but now for arrays
// CT = casa-type (casa::Int, casa::Double etc)
// AT = as_<intermediatetype> 'policy'; all
//     PyObjects will be first converted to
//     AT::return_type and then from
//     AT::return_type to CT
template <typename CT, typename AT>
struct ArrayFromPython {
    ArrayFromPython(PyObject* o) {
        Py_XINCREF(o);
        
        JPosition   ashp( sequencesupport::Shaper::get_shape(o) );

        // if the JPosition is not an empty one the shape
        // is 'well behaved'
        if( ashp==sequencesupport::noshape ) {
            Py_XDECREF(o);
            throw shape_of_subarrays_not_all_equal();
        }

        // now attempt the fill
        bool        fillok;

        // pre-resize the array
        value.resize( ashp );

        // and fill in the array
        // from the sequence-thingy 'o'
        fillok = sequencesupport::Basefiller::astype<AT>::do_fill(value, o);
        Py_XDECREF(o);
        if( !fillok ) {
            cerr << "Failed to warp PyObject into Array<"
                 << typeid(CT).name() << "> [may need c++filt to decode]"
                 << endl;
            throw incompatible_pythontype_for_columndatatype();
        }
    }
    casa::Array<CT>   value;
};

// proxy for a column-object
//   Easier to keep state 
struct ColumnProxy {

    ColumnProxy( const casa::Table& t, const std::string& col, bool af=false ):
        autoflush( af ),
        tab(t),
        colName(col)
    {
        if( !tab.tableDesc().isColumn(colName) )
            throw column_proxy_no_such_column();
        cd = t.tableDesc().columnDesc(colName);
    }

    unsigned int length( void ) const {
        return tab.nrow();
    }

    // access the cell at row 'idx' in the column
    // supports Python indexing scheme where
    // negative indices count backward from the end...
    PyObject* getCell( int idx ) {
        // If idx out of range, return None
        // but first see if we must adjust the idx
        if( idx<0 ) 
            idx += (int)tab.nrow();
        // if after adjustment, idx out of range
        // throw up
        if( idx<0 || idx>=tab.nrow() ) {
            PyErr_Format(PyExc_StopIteration, "");
            throw boost::python::error_already_set();
        }

        switch( cd.trueDataType() ) {
            case casa::TpBool: {
                    scolaccessor<casa::Bool>    b(tab, colName, idx);
                    return PyInt_FromLong( ((b.value==casa::True)?(1L):(0L)) );
                }
                break;
            case casa::TpArrayBool: {
                    const acolaccessor<casa::Bool>  d(tab, colName, idx);
                    return arraytransformer<warp::casabool2int>()( d.value );
                }
                break;
            case casa::TpInt: {
                    scolaccessor<casa::Int>    b(tab, colName, idx);
                    return PyInt_FromLong( (long)b.value );
                }
                break;
            case casa::TpArrayInt: {
                    const acolaccessor<casa::Int>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;

            case casa::TpString: {
                    scolaccessor<casa::String>  s(tab, colName, idx);
                    return PyString_FromString(s.value.chars());
                }
                break;
            case casa::TpArrayString: {
                    const acolaccessor<casa::String>  d(tab, colName, idx);
                    return arraytransformer<warp::ToPyStr>()( d.value );
                }
                break;

            case casa::TpFloat: {
                    scolaccessor<casa::Float>  s(tab, colName, idx);
                    return PyFloat_FromDouble((double)s.value);
                }
                break;
            case casa::TpArrayFloat: {
                    const acolaccessor<casa::Float>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;

            case casa::TpDouble: {
                    scolaccessor<casa::Double>  s(tab, colName, idx);
                    return PyFloat_FromDouble(s.value);
                }
                break;
            case casa::TpArrayDouble: {
                    const acolaccessor<casa::Double>  d(tab, colName, idx);
                    return arraytransformer<>()( d.value );
                }
                break;

            case casa::TpComplex: {
                    scolaccessor<casa::Complex>  s(tab, colName, idx);
                    return PyComplex_FromDoubles(s.value.real(), s.value.imag());
                }
                break;
            case casa::TpArrayComplex: {
                    const acolaccessor<casa::Complex>  d(tab, colName, idx);
                    return arraytransformer<warp::ToPyComplex>()( d.value );
                }
                break;

            case casa::TpDComplex: {
                    scolaccessor<casa::DComplex>  s(tab, colName, idx);
                    return PyComplex_FromDoubles(s.value.real(), s.value.imag());
                }
                break;
            case casa::TpArrayDComplex: {
                    const acolaccessor<casa::DComplex>  d(tab, colName, idx);
                    return arraytransformer<warp::ToPyDComplex>()( d.value );
                }
                break;

            default:
                break;
        }
        // Unsupported type
        throw column_proxy_unsupported_column_datatype();
    }

    // Now set the value of a given cell in row 'r'
    bool putCell( int r, PyObject* o ) {
        // If idx out of range, return None
        // but first see if we must adjust the idx
        if( r<0 ) 
            r += (int)tab.nrow();

        // if after adjustment, idx out of range
        // throw up
        if( r<0 || r>=tab.nrow() ) {
            PyErr_Format(PyExc_StopIteration, "");
            throw boost::python::error_already_set();
        }

        // note: the 'ScalarFromPython' c'tors throw up
        // if the conversion fails
        switch( cd.trueDataType() ) {
            case casa::TpBool: {
                    ScalarFromPython<warp::as_bool>  w(o);
                    scolputter<casa::Bool>(tab, colName, r, w.value);
                }
                break;

            case casa::TpString: {
                    ScalarFromPython<warp::as_stringexact>  w(o);
                    scolputter<casa::String>(tab, colName, r, w.value);
                }
                break;

            case casa::TpInt: {
                    ScalarFromPython<warp::as_int>  w(o);
                    scolputter<casa::Int>(tab, colName, r, w.value);
                }
                break;
            case casa::TpFloat: {
                    ScalarFromPython<warp::as_float>  w(o);
                    scolputter<casa::Float>(tab, colName, r, w.value);
                }
                break;
            case casa::TpDouble: {
                    ScalarFromPython<warp::as_double>  w(o);
                    scolputter<casa::Double>(tab, colName, r, w.value);
                }
                break;
            case casa::TpComplex: {
                    ScalarFromPython<warp::as_fcomplex>  w(o);
                    scolputter<casa::Complex>(tab, colName, r, casa::Complex(w.value.real(),
                                                                             w.value.imag()));
                }
                break;
            case casa::TpDComplex: {
                    ScalarFromPython<warp::as_dcomplex>  w(o);
                    scolputter<casa::DComplex>(tab, colName, r, casa::DComplex(w.value.real(),
                                                                               w.value.imag()));
                }
                break;
            // Array Stuff
            case casa::TpArrayBool: {
                    ArrayFromPython<casa::Bool,warp::as_bool>  w(o);
                    acolputter<casa::Bool>(tab, colName, r, w.value);
                }
                break;
            case casa::TpArrayInt: {
                    ArrayFromPython<casa::Int,warp::as_int>  w(o);
                    acolputter<casa::Int>(tab, colName, r, w.value);
                }
                break;
            case casa::TpArrayFloat: {
                    ArrayFromPython<casa::Float,warp::as_float>  w(o);
                    acolputter<casa::Float>(tab, colName, r, w.value);
                }
                break;
            case casa::TpArrayDouble: {
                    ArrayFromPython<casa::Double,warp::as_double>  w(o);
                    acolputter<casa::Double>(tab, colName, r, w.value);
                }
                break;
            case casa::TpArrayComplex: {
                    ArrayFromPython<casa::Complex,warp::as_fcomplex>  w(o);
                    acolputter<casa::Complex>(tab, colName, r, w.value);
                }
                break;
            case casa::TpArrayDComplex: {
                    ArrayFromPython<casa::DComplex,warp::as_dcomplex>  w(o);
                    acolputter<casa::DComplex>(tab, colName, r, w.value);
                }
                break;
            default:
                // Unsupported type
                throw column_proxy_unsupported_column_datatype();
                break;
        }
        // if we arrive here it means that no exceptions have been
        // thrown, ie stuff just worked!
        // flush the table to disk
        if( autoflush )
            tab.flush(casa::True);
        return true;
    }

    // get all the values in the column
    PyObject* get( void ) {
        switch( cd.trueDataType() ) {
            case casa::TpBool: {
                    const colaccessor2<casa::Bool>    b(tab, colName);
                    return arraytransformer<warp::casabool2int>()( b.values );
                }
                break;
            case casa::TpArrayBool: {
                    const colaccessor2<casa::Bool,casa::ROArrayColumn,casa::Array>   b(tab, colName);
                    return arraytransformer<warp::casabool2int>()( b.values );
                }
                break;

            case casa::TpInt: {
                    const colaccessor2<casa::Int>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;
            case casa::TpArrayInt: {
                    const colaccessor2<casa::Int,casa::ROArrayColumn,casa::Array>   b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case casa::TpFloat: {
                    const colaccessor2<casa::Float>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;
            case casa::TpArrayFloat: {
                    const colaccessor2<casa::Float,casa::ROArrayColumn, casa::Array>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case casa::TpDouble: {
                    const colaccessor2<casa::Double>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;
            case casa::TpArrayDouble: {
                    const colaccessor2<casa::Double,casa::ROArrayColumn, casa::Array>    b(tab, colName);
                    return arraytransformer<>()( b.values );
                }
                break;

            case casa::TpString: {
                    const colaccessor2<casa::String>     s(tab, colName);
                    return arraytransformer<warp::ToPyStr>()( s.values );
                }
                break;

            case casa::TpArrayComplex: {
                    const colaccessor2<casa::Complex, casa::ROArrayColumn, casa::Array>  d(tab, colName);
                    return arraytransformer<warp::ToPyComplex>()( d.values );
                }
                break;
            case casa::TpArrayDComplex: {
                    const colaccessor2<casa::DComplex, casa::ROArrayColumn, casa::Array>  d(tab, colName);
                    return arraytransformer<warp::ToPyDComplex>()( d.values );
                }
                break;
            default:
                // Unsupported type
                cerr << "pycasa: reading a column of this type (" 
                     << cd.trueDataType() << ") not yet supported."
                     << "mail verkouter@jive.nl if you want it added!"
                     << endl;
        }
        Py_INCREF(Py_None);
        return Py_None;
    }

    bool put( PyObject* o ) {
        switch( cd.trueDataType() ) {
            case casa::TpBool: {
                    ArrayFromPython<casa::Bool, warp::as_bool>  w(o);
                    colputter2<casa::Bool>                      b(tab, colName, w.value);
                }
                break;
            case casa::TpArrayBool: {
                    ArrayFromPython<casa::Bool, warp::as_bool>  w(o);
                    colputter2<casa::Bool, casa::ArrayColumn, casa::Array>  b(tab, colName, w.value);
                }
                break;

            case casa::TpInt: {
                    ArrayFromPython<casa::Int, warp::as_int>  w(o);
                    colputter2<casa::Int>                     b(tab, colName, w.value);
                }
                break;
            case casa::TpArrayInt: {
                    ArrayFromPython<casa::Int, warp::as_int>  w(o);
                    colputter2<casa::Int, casa::ArrayColumn, casa::Array> b(tab, colName, w.value);
                }
                break;

            case casa::TpFloat: {
                    ArrayFromPython<casa::Float, warp::as_float>  w(o);
                    colputter2<casa::Float>                       b(tab, colName, w.value);
                }
                break;
            case casa::TpArrayFloat: {
                    ArrayFromPython<casa::Float, warp::as_float>  w(o);
                    colputter2<casa::Float, casa::ArrayColumn, casa::Array>   b(tab, colName, w.value);
                }
                break;

            case casa::TpDouble: {
                    ArrayFromPython<casa::Double, warp::as_double>  w(o);
                    colputter2<casa::Double>                        b(tab, colName, w.value);
                }
                break;
            case casa::TpArrayDouble: {
                    ArrayFromPython<casa::Double, warp::as_double>  w(o);
                    colputter2<casa::Double, casa::ArrayColumn, casa::Array>    b(tab, colName, w.value);
                }
                break;

            case casa::TpString: {
                    ArrayFromPython<casa::String, warp::as_stringexact>  w(o);
                    colputter2<casa::String>                        b(tab, colName, w.value);
                }
                break;

            case casa::TpArrayComplex: {
                    ArrayFromPython<casa::Complex, warp::as_fcomplex>  w(o);
                    colputter2<casa::Complex, casa::ArrayColumn, casa::Array>      b(tab, colName, w.value);
                }
                break;
            case casa::TpArrayDComplex: {
                    ArrayFromPython<casa::DComplex, warp::as_dcomplex>  w(o);
                    colputter2<casa::DComplex, casa::ArrayColumn, casa::Array>      b(tab, colName, w.value);
                }
                break;
            default:
                // Unsupported type
                cerr << "pycasa: writing a column of this type (" 
                     << cd.trueDataType() << ") not yet supported."
                     << "mail verkouter@jive.nl if you want it added!"
                     << endl;
                return false;
        }
        // if we arrive here it means that no exceptions have been
        // thrown, ie stuff just worked!
        // flush the table to disk
        if( autoflush )
            tab.flush(casa::True);
        return true;
    }


    // get the column keywords
    PyObject* keywordSet( void ) {
        return boost::python::to_python_value<casa::RecordInterface const&>()( cd.keywordSet() );
    }

    bool              autoflush;
    casa::Table       tab;
    casa::ColumnDesc  cd;
    std::string colName;

    private:
        ColumnProxy();
};

std::ostream& operator<<( std::ostream& os, const ColumnProxy& cp ) {
    return os << cp.cd;
}


//
// Proxy for a casa Table
//    very thin wrapper
//    but makes it easy for us
//    and Boost.Python to bind
//    only specific members and/or
//    do some processing before/after
//    the actual call to the implementation
struct TableProxy {

    // default c'tor
    TableProxy():
        strict( true ), autoflush( false ),
        tab(), ra( tab )
    {}

    // do implement the copy c'tor ->
    // this'un gets called by the BOOST stuff
    // and we must take precautions to 
    // make sure the (refcounted) table
    // and ROTableRow are properly initialized...
    // 
    // It certainly isn't valid to "just copy
    // other.tab and other.ra across" even though
    // both objects support copy construction.
    //
    // The problem is with the ROTableRow, it is,
    // when it is constructed, associated with a
    // specific table. The ROTableRow in a TableProxy
    // should always be 'associated' with the table
    // object in the same object.
    //
    // In the example below, without a "proper" (read: correct)
    // copy c'tor, after executing these statements,
    // TableProxy b's ROTableRow member references
    // TableProxy a's Table and *not* b's Table,
    // which messes up reference counting [which happens
    // inside the Tables] and hence cleaning up of
    // tempfiles etc.
    //
    // TableProxy    a;
    // TableProxy    b(a);
    //
    // A 'correct' copy constructor is implemented below,
    // we can just copy construct our table with other's table
    // but we must explicitly associate our ROTableRow with
    // *our* table-object
    TableProxy( const TableProxy& other ):
        strict( other.strict ), autoflush( other.autoflush ),
        tab( other.tab ), ra( tab )
    {}

    const TableProxy& operator=( const TableProxy& other ) {
        if( this!=&other ) {
            strict    = other.strict;
            autoflush = other.autoflush;
            tab       = other.tab;
            ra        = casa::ROTableRow(tab);
        }
        return *this;
    }

    TableProxy( const casa::Table& t ):
        strict( true ), autoflush( false ),
        tab( t ), ra( tab )
    {}

    // default to 'read-only'. If ro==false => opened with
    // Table::Update
    TableProxy( const std::string& s, bool ro=true, bool st=true, bool af=false ):
        strict( st ), autoflush( af ),
        tab( s, ((ro==true)?(casa::Table::Old):(casa::Table::Update)) ), ra( tab )
    {}
#if 0
    typedef boost::python::detail::keywords<2>  tp_kwargs_type;
    TableProxy( const std::string& s, const tp_kwargs_type& kwargs = tp_kwargs_type() ) {
        cout << "TableProxy(" << s << ", **kwargs" << endl;
    }
#endif
#if 0
    TableProxy( const std::string& s, PyObject* kwargs=0 ) {
        Table::TableOption   to( Table::Old );
        // See what's in the record, if anything
        if( kwargs!=0 ) {
            if( !PyDict_Check(kwargs) )
                throw method_accepts_only_kwargs();
            // accepted keys
            static PyObject*  ro_key = PyString_FromString("ro");
            static PyObject*  af_key = PyString_FromString("autoflush");

            // see if the 'ro=<something>' key was given
//            if( PyDict_Contains(kwargs, ro_key) ) {
            if( PyDict_GetItem(kwargs, ro_key)!=0 ) {
                warp::as_bool::return_type    roflag;

                if( !warp::as_bool::do_convert(roflag, PyDict_GetItem(kwargs, ro_key)) ) {
                    cerr << "TableProxy/value of key 'ro' is not convertible to bool" << endl;
                    throw incompatible_pythontype_for_requesteddatatype();
                }
                if( roflag==true )
                    to = Table::Old;
                else
                    to = Table::Update;
            }
//            if( PyDict_Contains(kwargs, af_key) ) {
            if( PyDict_GetItem(kwargs, af_key)!=0 ) {
                warp::as_bool::return_type    afflag;

                if( !warp::as_bool::do_convert(afflag, PyDict_GetItem(kwargs, af_key)) ) {
                    cerr << "TableProxy/value of key 'autoflush' is not convertible to bool" << endl;
                    throw incompatible_pythontype_for_requesteddatatype();
                }
                cerr << "TableProxy/Setting autoflush to " << afflag << endl;
            }
        }
        // done processing the kwargs, no create you b****!
        tab = Table(s, to);
        ra  = ROTableRow(tab);
    }
#endif
    // attempt to re-open the table for updating.
    // The .reopenRW() method is called and 
    // the doc. sais that an exception is thrown
    // if that fails. We catch the exception and
    // transform that into a false return value.
    bool update( void ) {
        bool   rv( true );
        try {
            tab.reopenRW();
        }
        catch( ... ) {
            rv = false;
        }
        return rv;
    }
    // attempt to unlock the table
    // [if PermanentLock is present, this will not
    //  unlock that, obviously. It's just that I don't
    //  know if/when/how PermaLocks are set...]
    void unlock( void ) {
        tab.unlock();
        return;
    }
    unsigned int ncol( void ) const {
        return tab.tableDesc().ncolumn();
    }

    unsigned int nrow( void ) const {
        return tab.nrow();
    }

    // and make a deepcopy of the current table!
    void deepcopy( const std::string& newtabname ) {
        tab.deepCopy( newtabname, casa::Table::NewNoReplace, casa::True ); 
        return;
    }

    // all the columnnames as array of strings
    PyObject* colnames( void ) const {
        casa::Vector<casa::String>  cn( tab.tableDesc().columnNames() );
        return arraytransformer<warp::ToPyStr>()((const casa::Array<casa::String>&)cn);
    }
   
    // get all the values in row 'idx' in the table
    // supports python-style indexing where negative
    // indices count backwards from the end, ie, backward
    // from the last row of the table...
    PyObject* getRow( int r ) {
        static  boost::python::to_python_value<casa::RecordInterface const&>  warper;
        if( r<0 ) 
            r += (int)tab.nrow();

        if( r<0 || r>=(int)tab.nrow() ) {
            PyErr_Format(PyExc_StopIteration, "");
            throw boost::python::error_already_set();
        }
        return warper(ra.get(r));
    }
    // accept a dict as:
    //   table[n] = {'colname1':<value1>, ..., 'colnameN':<valueN>}
    bool putRow( int r, PyObject* o ) {
        if( !PyDict_Check(o) ) {
            PyErr_SetString(PyExc_ValueError,"Argument must be dict");
            throw boost::python::error_already_set();
        }

        // will be set to false in case of caught exception
        bool    rv( true );
        
        try {
            // if strict mode, verify that all keys() in the dict
            // are valid columnnames
            if( strict ) {
                PyObject*                                     keyobj;
                ArrayFromPython<casa::String,warp::as_string> skeys( PyDict_Keys(o) );
    
                if( skeys.value.shape().nelements()!=1 ) {
                    cout << "Huh? Multidimensional string array for dict.keys()?"
                         << "expected ndim==1" << endl;
                    return false;
                }
                casa::Int                   nkeys;
                const casa::TableDesc&      td( tab.tableDesc() );
                casa::Vector<casa::String>  keyv( skeys.value );
            
                keyv.shape( nkeys );
                for( casa::Int ki=0; ki<nkeys; ++ki ) {
                    if( !td.isColumn( keyv[ki] ) ) {
                        cout << "Table::putRow()/strict mode && table has no column "
                             << "named '" << keyv[ki] << "'" << endl;
                        return false;
                    }
                }
            }

            // Now iterate over all the K,V pairs and set the column
            // values
            int       ki;
            bool      oldautoflush = autoflush;
            PyObject* keyptr;
            PyObject* valueptr;

            // turn of autoflush - we will do that
            // after we've updated all columns
            ki        = 0;
            rv        = true;
            autoflush = false;

            while( PyDict_Next(o, &ki, &keyptr, &valueptr) ) {
                ColumnProxy*                        colproxy;
                ScalarFromPython<warp::as_string>   key( keyptr );

                colproxy = this->getColumn( key.value );
                try {
                    rv = rv && colproxy->putCell(r, valueptr );
                }
                catch( const std::exception& e ) {
                    rv = false;
                    cout << "Caugth a " << typeid(e).name() << " exception " << endl;
                }
                catch( ... ) {
                    rv = false;
                    cout << "Caught unknown exception " << endl;
                }
                delete colproxy;
                if( !rv ) {
                    cout << "Table::putRow()/Failed to set value of "
                         << key.value << " at row #" << r << endl;
                    if( strict )
                        return false;
                    else
                        cout << "Continuing with other fields (non-strict mode)" << endl;
                }
            }
            autoflush = oldautoflush;
            if( autoflush )
                this->flush();
        }
        catch( std::exception& e ) {
            rv = false;
            cout << "Table::putRow()/Got an exception: "
                 << typeid(e).name()
                 << endl;
        }
        return rv;
    }

    ColumnProxy* getColumn( const std::string& colnm ) {
        return new ColumnProxy(tab, colnm, autoflush);
    }

    const casa::String& tableName( void ) const {
        return tab.tableName();
    }

    PyObject* keywordSet( void ) {
        return boost::python::to_python_value<casa::RecordInterface const&>()( tab.keywordSet() );
    }

    void flush( void ) {
        tab.flush( casa::True );
        return;
    }

    TableProxy  query( const std::string& taql ) {
        return TableProxy( casa::tableCommand(casa::String("select from $1 where ")+casa::String(taql), tab) );

    }

    // query with select clause
    TableProxy query( const std::string& cols, const std::string& where ) {
        casa::String  qry("select ");

        if( cols.size() )
            qry += casa::String(cols);
        qry += " from $1 ";
        if( where.size() )
            qry += casa::String("where ")+casa::String(where);
        return TableProxy( casa::tableCommand(qry, tab) );
    }

    // taql on this table. your expression
    // is sent throught TaQL as-is
    TableProxy taql( const std::string& s ) {
        return TableProxy( casa::tableCommand(casa::String(s), tab) );
    }

    bool              strict;
    bool              autoflush;
    casa::Table       tab;
    casa::ROTableRow  ra;
};

// wrapper so as we do not have to bother about overloaded
// tableCommand() fn's. This is just the one taking a single
// string as arg, the full TaQL query
TableProxy  tc_wrapper(const std::string& s) {
    return TableProxy( casa::tableCommand(casa::String(s)) );
}

// Create function pointers to both versions of TableProxy::query
// in order to help Boost decide which to select...
TableProxy (TableProxy::*q1)(const std::string&) = &TableProxy::query;
TableProxy (TableProxy::*q2)(const std::string&, const std::string&) = &TableProxy::query;

PyObject* shapetst( PyObject* o ) {
    Py_XINCREF(o);
    casa::IPosition           shp( sequencesupport::Shaper::get_shape(o) );
    Py_XDECREF(o);
    return arraytransformer<>()( shp.asVector() );
}

PyObject* arraytst( PyObject* o ) {
    // attempt to convert to array of ints
    Py_XINCREF(o);
    JPosition           shp( sequencesupport::Shaper::get_shape(o) );
    Py_XDECREF(o);
    if( shp!=sequencesupport::noshape ) {
        casa::Array<casa::Int>  arr( shp );
        cerr << "Warping pyobject into array<Int> of shape " << shp << endl;
        if( sequencesupport::Basefiller::astype<warp::as_int>::do_fill(arr, o) ) {
            return arraytransformer<>()( arr );
        }
    }
    cerr << "arraytest: failed to warp PyObject* into Array<Int>" << endl;
    Py_INCREF(Py_None);
    return Py_None;
}

PyObject* strtst( PyObject* o ) {
    // attempt to convert to array of strings
    Py_XINCREF(o);
    PyObject*           rv( 0 );
    JPosition           shp( sequencesupport::Shaper::get_shape(o) );
    if( shp!=sequencesupport::noshape ) {
        casa::Array<casa::String>  arr( shp );
        cerr << "Warping pyobject into array<String> of shape " << shp << endl;
        if( sequencesupport::Basefiller::astype<warp::as_string>::do_fill(arr, o) ) {
            rv = arraytransformer<warp::ToPyStr>()( arr );
        }
    }
    Py_XDECREF(o);
    if( !rv ) {
        cerr << "arraytest: failed to warp PyObject* into Array<String>" << endl;
        Py_INCREF(Py_None);
        rv = Py_None;
    }
    return rv;
}
PyObject* flttst( PyObject* o ) {
    // attempt to convert to array of Floats
    Py_XINCREF(o);
    PyObject*           rv( 0 );
    JPosition           shp( sequencesupport::Shaper::get_shape(o) );
    if( shp!=sequencesupport::noshape ) {
        casa::Array<casa::Float>  arr( shp );
        cerr << "Warping pyobject into array<Float> of shape " << shp << endl;
        if( sequencesupport::Basefiller::astype<warp::as_float>::do_fill(arr, o) ) {
            rv = arraytransformer<>()( arr );
        }
    }
    Py_XDECREF(o);
    if( !rv ) {
        cerr << "arraytest: failed to warp PyObject* into Array<Float>" << endl;
        Py_INCREF(Py_None);
        rv = Py_None;
    }
    return rv;
}
PyObject* cplxtst( PyObject* o ) {
    // attempt to convert to array of Floats
    Py_XINCREF(o);
    PyObject*           rv( 0 );
    JPosition           shp( sequencesupport::Shaper::get_shape(o) );
    if( shp!=sequencesupport::noshape ) {
        casa::Array<casa::Complex>  arr( shp );
        cerr << "Warping pyobject into array<Complex> of shape " << shp << endl;
        if( sequencesupport::Basefiller::astype<warp::as_fcomplex>::do_fill(arr, o) ) {
            rv = arraytransformer<>()( arr );
        }
    }
    Py_XDECREF(o);
    if( !rv ) {
        cerr << "arraytest: failed to warp PyObject* into Array<Complex>" << endl;
        Py_INCREF(Py_None);
        rv = Py_None;
    }
    return rv;
}
PyObject* booltst( PyObject* o ) {
    // attempt to convert to array of bools
    Py_XINCREF(o);
    PyObject*           rv( 0 );
    JPosition           shp( sequencesupport::Shaper::get_shape(o) );
    if( shp!=sequencesupport::noshape ) {
        casa::Array<casa::Bool>  arr( shp );
        cerr << "Warping pyobject into array<Bool> of shape " << shp << endl;
        if( sequencesupport::Basefiller::astype<warp::as_bool>::do_fill(arr, o) ) {
            rv = arraytransformer<warp::casabool2int>()( arr );
        }
    }
    Py_XDECREF(o);
    if( !rv ) {
        cerr << "arraytest: failed to warp PyObject* into Array<Bool>" << endl;
        Py_INCREF(Py_None);
        rv = Py_None;
    }
    return rv;
}

const char* pycasa_version( void ) {
    return VERS;
}

//
// The casa module
// 
BOOST_PYTHON_MODULE(pycasa)
{
    // define a table
    class_<TableProxy>("table")
        .def(init<>())
        .def(init<std::string>())
//        .def(init<std::string,optional<TableProxy::tp_kwargs_type> >())
        .def(init<std::string,optional<bool,bool,bool> >())
//        .def(init<std::string,PyObject*>())
        .def("tableName", &TableProxy::tableName,
                          return_value_policy<copy_const_reference>())
        .def("__len__", &TableProxy::nrow)
        .def("ncol", &TableProxy::ncol)
        .def("getcol", &TableProxy::getColumn,
                       return_internal_reference<>())
        .def("__getitem__", &TableProxy::getRow)
        .def("__setitem__", &TableProxy::putRow)
        .def("query", q1)
        .def("query", q2)
        .def("taql", &TableProxy::taql)
        .def("colnames", &TableProxy::colnames)
        .def("keywordSet", &TableProxy::keywordSet)
        .def("update", &TableProxy::update)
        .def("flush", &TableProxy::flush)
        .def("deepcopy", &TableProxy::deepcopy)
        .def("unlock", &TableProxy::unlock)
        ;
        
    class_<ColumnProxy>("column", no_init)
        .def("__str__", &stringifier<ColumnProxy>::asString )
        .def("__getitem__", &ColumnProxy::getCell )
        .def("__setitem__", &ColumnProxy::putCell )
        .def("get", &ColumnProxy::get)
        .def("put", &ColumnProxy::put)
        .def("__len__",&ColumnProxy::length)
        .def("keywordSet", &ColumnProxy::keywordSet)
        ;

    // define the global taql function
    def("taql", &tc_wrapper);


    def("shapetst", &shapetst);
    def("inttst", &arraytst);
    def("strtst", &strtst);
    def("flttst", &flttst);
    def("cplxtst", &cplxtst);
    def("booltst", &booltst);
    def("version", &pycasa_version);

    // Do register type conversions
    CasaStringFromPythonString();
}

//
// Transform a recordinterface based field into a PyDict
// Need this wrapper since it must be defined *after*
// the boost::python::to_python_value<RecordInterface> is defined
//
PyObject* make_pydict_fromsubrecord(casa::RecordInterface const& r, casa::Int num) {
    // retrieve pointer to the subrecord
    const casa::RecordInterface& srec = *((casa::RecordInterface*)(r.get_pointer(num, casa::TpRecord)));

    // and warp it into a PyObject
    return boost::python::to_python_value<casa::RecordInterface const&>()(srec);
}

// Define this only after the class_<TableProxy> is defined 
// such that we can actually give that back
PyObject* make_pyobject_fromsubtable(casa::RecordInterface const& r, casa::Int num) {
    // retrieve pointer to the subtable
    const casa::TableKeyword& trec = *((const casa::TableKeyword*)(r.get_pointer(num, casa::TpTable)));
    // Prepare return value, it's going to be one of our own defined
    // python classes (the tableproxy thingy) and we get a reference to the
    // held C++ object of type 'TableProxy'. Then we set the value to the
    // new table and return the python object
    object  rv( new TableProxy(trec.table()) );
    
    incref( rv.ptr() );
    return rv.ptr();
}
