// Bind (selected) casa::Quanta functionality to Python
#include <Python.h>
#include <boost/python.hpp>

// c++ stuff
#include <sstream>

// Casa (http://casa.nrao.edu) specific
#include <casa/aips.h>
#include <casa/Quanta.h>
#include <casa/Quanta/QMath.cc>
#include <casa/Quanta/MVTime.h>
#include <casa/Quanta/MVAngle.h>
#include <casa/Quanta/UnitMap.h>
#include <casa/Quanta/QLogical.h>


using namespace std;
using namespace boost::python;
using namespace casa;




// Wrap the Quantity so we can insert some glue code between
// the python representation and the actual c++ thingy
struct MYQuantity:
    public Quantity
{
    MYQuantity():
        Quantity()
    {}

    // for copy.deepcopy
    MYQuantity deepcopy( PyObject* ) const {
        MYQuantity   rv( *this );
        return rv;
    }

    MYQuantity(double d):
        Quantity(d)
    {}

    MYQuantity(double d, const string& unit):
        Quantity(d, Unit(unit))
    {}

    // Copy + optional conversion
    // for both Quantity and self.
    // If 'unit' is not the empty string, "this" will
    // be set to "q converted to unit 'unit'"
    MYQuantity( const Quantity& q ):
        Quantity( q )
    {}

    MYQuantity(const Quantity& q, const string& unit) {
        if( unit.size() ) {
            Unit     u;
            UnitName tnam;
            // look it up in the UnitMap?
            if( UnitMap::getUnit(unit, tnam) ) {
                // yup
                u.setValue( tnam.getVal() );
                u.setName( tnam.getName() );
            } else {
                u = Unit(unit);
            }
            this->Quantity::operator=( q.get(u) );
        } else {
            this->Quantity::operator=(q);
        }
    }

    // From MYQuantity plus optional conversion
    MYQuantity( const MYQuantity& other ):
        Quantity( (const Quantity&)other)
    {}
    MYQuantity(const MYQuantity& q, const string& unit) {
        if( unit.size() ) {
            Unit     u;
            UnitName tnam;
            // look it up in the UnitMap?
            if( UnitMap::getUnit(unit, tnam) ) {
                // yup
                u.setValue( tnam.getVal() );
                u.setName( tnam.getName() );
            } else {
                u = Unit(unit);
            }
            this->Quantity::operator=( q.get(u) );
        } else {
            this->Quantity::operator=(q);
        }
    }
    // Construct from string
    MYQuantity( const string& s ) {
        String cs( s );
        if( !Quantity::read(*this, s) ) {
            PyErr_SetString(PyExc_ValueError, "Unsupported quantity?");
            throw boost::python::error_already_set();
        }
    }

    // string representation
    string  str( void ) const {
        ostringstream oss;
        this->print(oss);
        return oss.str();
    }

    // Conversion to different unit
    MYQuantity convert( const string& unit ) const {
        return MYQuantity(*this, unit);
    }

    // Arithmetic

    // Addition - note: there are some
    // specializations of those below!

    // "quantity + <something>" (Python: __add__)
    template <typename T>
    MYQuantity operator+( const T& other ) const {
        MYQuantity  rv( *this );

        rv.Quantity::operator+=( other );
        return rv;
    }
    // introduce this memberfunction to be able
    // to define __radd__ in order to support
    // "<built-in> + quantity"
    template <typename T>
    MYQuantity qadd(const T& l) const {
        return (*this)+l;
    }
    // In-place addition (Python: __iadd__)
    // also specialized for string
    template <typename T>
    const MYQuantity& operator+=( const T& t ) {
        this->Quantity::operator+=(t);
        return *this;
    }

    // Subtraction - note: there are some
    // specializations of those below!

    // "quantity - <something>" (Python: __sub__)
    template <typename T>
    MYQuantity operator-( const T& other ) const {
        MYQuantity  rv( *this );

        rv.Quantity::operator-=( other );
        return rv;
    }
    // introduce this memberfunction to be able
    // to define __rsub__ in order to support
    // "<built-in> - quantity"
    template <typename T>
    MYQuantity qsub(const T& l) const {
        return (*this)-l;
    }
    // In-place addition (Python: __isub__)
    // also specialized for string
    template <typename T>
    const MYQuantity& operator-=( const T& t ) {
        this->Quantity::operator-=(t);
        return *this;
    }

    // Multiplies - note: there are some
    // specializations of those below!

    // "quantity * <something>" (Python: __mul__)
    template <typename T>
    MYQuantity operator*( const T& other ) const {
        MYQuantity  rv( *this );

        rv.Quantity::operator*=( other );
        return rv;
    }
    // introduce this memberfunction to be able
    // to define __rmul__ in order to support
    // "<built-in> * quantity"
    template <typename T>
    MYQuantity qmul(const T& l) const {
        return (*this)*l;
    }
    // In-place addition (Python: __imul__)
    // also specialized for string
    template <typename T>
    const MYQuantity& operator*=( const T& t ) {
        this->Quantity::operator*=(t);
        return *this;
    }

    // Divides - note: there are some
    // specializations of those below!

    // "quantity / <something>" (Python: __div__)
    template <typename T>
    MYQuantity operator/( const T& other ) const {
        MYQuantity  rv( *this );

        rv.Quantity::operator/=( other );
        return rv;
    }
    // introduce this memberfunction to be able
    // to define __rdiv__ in order to support
    // "<built-in> / quantity"
    template <typename T>
    MYQuantity qdiv(const T& l) const {
        return (*this)/l;
    }
    // In-place addition (Python: __idiv__)
    // also specialized for string
    template <typename T>
    const MYQuantity& operator/=( const T& t ) {
        this->Quantity::operator/=(t);
        return *this;
    }


    // Comparison
    // Only define operator< for specific types.
    // All other relational operators have been rewritten
    // to only use "operator<".
    //
    // Thus, as long as the templates
    // convert unknown things to things that we know
    // of how to compare we're ok and we support that
    // specific comparison!
    bool operator<(const MYQuantity& other) const {
        return ((const Quantity&)*this) < ((const Quantity&)other);
    }
    // first try conversion to MYQuantity, *then* do the compare
    template <typename T>
    bool operator<(const T& t) const {
        return *this<MYQuantity(t);
    }
    template <typename T>
    bool operator<=(const T& t) const {
        const MYQuantity other(t);
        return (*this<other || (other<*this)==false);
    }
    template <typename T>
    bool operator==(const T& t) const {
        const MYQuantity other( t );
        return (*this<other)==false && (other<*this)==false;
    }
    template <typename T>
    bool operator!=(const T& t) const {
        const MYQuantity other( t );
        return (*this<other || other<*this);
    }
    template <typename T>
    bool operator>(const T& t) const {
        const MYQuantity other( t );
        return other<*this;
    }
    template <typename T>
    bool operator>=(const T& t) const {
        const MYQuantity other( t );
        return (other<*this || (*this<other)==false);
    }
};
// make specialized versions for string!
// add
template <>
MYQuantity MYQuantity::operator+( const string& s ) const {
    return (*this)+MYQuantity(s);
}
template <>
MYQuantity MYQuantity::qadd( const string& s ) const {
    return (*this)+MYQuantity(s);
}
template <>
const MYQuantity& MYQuantity::operator+=( const string& s ) {
    this->operator+=( MYQuantity(s) );
    return *this;
}

// sub
template <>
MYQuantity MYQuantity::operator-( const string& s ) const {
    return (*this)-MYQuantity(s);
}
template <>
MYQuantity MYQuantity::qsub( const string& s ) const {
    return (*this)-MYQuantity(s);
}
template <>
const MYQuantity& MYQuantity::operator-=( const string& s ) {
    this->operator-=( MYQuantity(s) );
    return *this;
}

// mul
template <>
MYQuantity MYQuantity::operator*( const string& s ) const {
    return (*this)*MYQuantity(s);
}
template <>
MYQuantity MYQuantity::qmul( const string& s ) const {
    return (*this)*MYQuantity(s);
}
template <>
const MYQuantity& MYQuantity::operator*=( const string& s ) {
    this->operator*=( MYQuantity(s) );
    return *this;
}
// div
template <>
MYQuantity MYQuantity::operator/( const string& s ) const {
    return (*this)/MYQuantity(s);
}
template <>
MYQuantity MYQuantity::qdiv( const string& s ) const {
    return (*this)/MYQuantity(s);
}
template <>
const MYQuantity& MYQuantity::operator/=( const string& s ) {
    this->operator/=( MYQuantity(s) );
    return *this;
}

// I'm afraid we'll have to wrap the MVTime as well
// (if only to support construction from a string!)
struct MYTime:
    public MVTime
{
    MYTime():
        MVTime()
    {}

    MYTime( const MYTime& other ):
        MVTime( (const MVTime&)other )
    {}

    // gets passed the Memo dict - discard it for now
    MYTime  deepcopy( PyObject* ) const {
        MYTime   rv( *this );
        return rv;
    }

    // construct from a Quantity
    MYTime( const MYQuantity& other ):
        MVTime( (const Quantity&)other )
    {}

    // construct from a string - should allow all
    // formats that casa groks, plus add a few (one?)
    // of our own. Also try different things...
    MYTime( const std::string& tstr ):
        MVTime()
    {
        String     cs( tstr );
        Quantity   qtmp;
       
        // First try MVTime::read()
        if( !MVTime::read(qtmp, cs) ) {
            double         s;
            unsigned int   h, m;
            // Try vanilla time format "h:m:s" with h,m integers
            // and s being floating point [so supports fractional seconds]
            // today/0:0:0 will be automagically added. Only times between
            // 0:0:0 and 23:59:59.99999 [the number of 9's behind the decimal
            // point was arbitrarily chosen]
            if( ::sscanf(cs.c_str(), "%u:%u:%lf", &h, &m, &s)==3 ) {
                // Check if it is just "h:m:s" format 
                // (since casa::MVTime doesn't support it -
                //  we would like it though)
                double         d;
                const double   slim( 59.99999 );
                const MYTime   today("today/0:0:0");

                // Check for sane values
                if( h>23 || m>59 || s>slim ) {
                    ostringstream e;

                    e << "Time out of range either h>24, m>59 or s>" << slim;
                    PyErr_SetString(PyExc_ValueError, e.str().c_str());
                    throw boost::python::error_already_set();
                }
                // Now convert to fractional day (or days) and add
                // "today" to it.
                d = ((double)h/24.0) + ((double)m/(24.0*60.0)) +
                    s/(24.0*60.0*60.0);
                qtmp = Quantity(today.get("d")+Quantity(d, "d"));
            } else {
                // We can always try Quantity::read() ...
                if( !Quantity::read(qtmp, cs) ) {
                    // even that failed... Now we finally give up
                    PyErr_SetString(PyExc_ValueError, "Unsupported timeformat");
                    throw boost::python::error_already_set();
                }
            }
        }
        // If we arrive here, qtmp has been filled in with something
        // so construct a time from it
        this->MVTime::operator=( MVTime(qtmp) );
    }

    // I think I'd like this to return a full time... not
    // just "h:m:s" (especially since you have *no idea* on
    // which day/year/month that time is...
    std::string str( void ) const {
        return this->format( MYTime::defStrFormat );
    }

    // This one will be overloaded using BOOST_PYTHON_MEMBER_FUNCTION_OVERLOAD
    // below
    std::string format(MVTime::formatTypes fmt, unsigned int inprec=0) const {
        return this->MVTime::string(fmt, inprec);
    }

    // Get the time, optionally in specified unit.
    MYQuantity  get( const std::string& unit="" ) const {
        // If empty unitstring -> return the default ".get()" method
        if( unit.empty() )
            return this->MVTime::get();
       
        // Attempt to convert to requested unit
        Unit     u;
        UnitName tnam;

        // look unit up in the UnitMap?
        if( UnitMap::getUnit(unit, tnam) ) {
            // yup
            u.setValue( tnam.getVal() );
            u.setName( tnam.getName() );
        } else {
            u = Unit(unit);
        }
        return this->MVTime::get(u);
    }

    // Comparison
    bool operator<(const MYTime& other) const {
        return (Double)(*this)<(Double)other;
    }
    bool operator<(const MYQuantity& other) const {
        return (*this)<MYTime((const Quantity&)other);
    }

    // Same scheme as with the Quantity
    // first try conversion to MYTime, *then* do the compare
    template <typename T>
    bool operator<(const T& t) const {
        return *this<MYTime(t);
    }
    template <typename T>
    bool operator<=(const T& t) const {
        const MYTime other(t);
        return (*this<other || (other<*this)==false);
    }
    template <typename T>
    bool operator==(const T& t) const {
        const MYTime other( t );
        return (*this<other)==false && (other<*this)==false;
    }
    template <typename T>
    bool operator!=(const T& t) const {
        const MYTime other( t );
        return (*this<other || other<*this);
    }
    template <typename T>
    bool operator>(const T& t) const {
        const MYTime other( t );
        return other<*this;
    }
    template <typename T>
    bool operator>=(const T& t) const {
        const MYTime other( t );
        return (other<*this || (*this<other)==false);
    }

    // Extend the MVTime with addition and subtraction

    // Define it for known types
    const MYTime& operator+=( const MYTime& t ) {
        this->MVTime::operator=( MVTime(this->get()+t.get()) );
        return *this;
    }
    // Now we can throw in the templates
    template <typename T>
    const MYTime& operator+=(const T& t) {
        // Support anything we can convert to a MYTime...
        (*this)+=MYTime(t);
        return *this;
    }

    // Define the operator+ in terms of += (so we only have to
    // maintain the += operators)
    template <typename T>
    MYTime operator+( const T& t ) const {
        MYTime     rv( *this );
        rv+=t;
        return rv;
    }

    // subtraction
    
    // Define the inplace 'operator-' (-=) for 
    // specific types
    const MYTime& operator-=( const MYTime& t ) {
        this->MVTime::operator=( MVTime(this->get()-t.get()) );
        return *this;
    }
    // throw in the templates and the operator-
    template <typename T>
    const MYTime& operator-=(const T& t) {
        // Support anything we can convert to MYTime...
        (*this)-=MYTime(t);
        return *this;
    }

    // Define operator- in terms of 'operator-='
    template <typename T>
    MYTime operator-(const T& t) {
        MYTime  rv( *this );
        rv-=t;
        return rv;
    }

    // Remember a default type?
    static MVTime::formatTypes   defStrFormat;
};

// Support builtin+MYTime
template <typename T>
MYTime mytimesub(const T& t, const MYTime& r) {
    MYTime    l( t );
    l-=r;
    return l;
}

MVTime::formatTypes MYTime::defStrFormat = MVTime::DMY;

// Overload the get method from the MYTime thingy
BOOST_PYTHON_MEMBER_FUNCTION_OVERLOADS(mytimeget, get, 0, 1);

// Overload the format method from the MYTime thingy
BOOST_PYTHON_MEMBER_FUNCTION_OVERLOADS(mytimeformat, format, 1, 2);

// For specifiying in second argument in comparison:
typedef other<MYQuantity>   OQuant;

// The pyquanta module
BOOST_PYTHON_MODULE(pyquanta) {
    // The Quantity
    class_<MYQuantity>("quantity")
        .def(init<>())
        .def(init<double>())
        .def(init<double,const string&>())
        .def(init<const string&>())
        .def(init<const Quantity&>())
        .def(init<const MYQuantity&>())
        .def(init<const Quantity&, const string&>())
        .def(init<const MYQuantity&, const string&>())
        .def("__deepcopy__", &MYQuantity::deepcopy)
        .def("get", &MYQuantity::convert)
        .def("__str__", &MYQuantity::str)
        .def("__repr__", &MYQuantity::str)
        // add
        .def(self+self)
        .def(self+double())
        .def(self+int())
        .def(self+string())
        .def("__radd__", &MYQuantity::qadd<int>)
        .def("__radd__", &MYQuantity::qadd<double>)
        .def("__radd__", &MYQuantity::qadd<string>)
        .def(self+=int())
        .def(self+=double())
        .def(self+=string())
        // sub
        .def(self-self)
        .def(self-double())
        .def(self-int())
        .def(self-string())
        .def("__rsub__", &MYQuantity::qsub<int>)
        .def("__rsub__", &MYQuantity::qsub<double>)
        .def("__rsub__", &MYQuantity::qsub<string>)
        .def(self-=int())
        .def(self-=double())
        .def(self-=string())
        // mul
        .def(self*self)
        .def(self*double())
        .def(self*int())
        .def(self*string())
        .def("__rmul__", &MYQuantity::qmul<int>)
        .def("__rmul__", &MYQuantity::qmul<double>)
        .def("__rmul__", &MYQuantity::qmul<string>)
        .def(self*=int())
        .def(self*=double())
        .def(self*=string())
        // div
        .def(self/self)
        .def(self/double())
        .def(self/int())
        .def(self/string())
        .def("__rdiv__", &MYQuantity::qdiv<int>)
        .def("__rdiv__", &MYQuantity::qdiv<double>)
        .def("__rdiv__", &MYQuantity::qdiv<string>)
        .def(self/=int())
        .def(self/=double())
        .def(self/=string())
        // Comparison
        .def(self<self)
        .def(self<=self)
        .def(self==self)
        .def(self!=self)
        .def(self>self)
        .def(self>=self)
        // compare to double
        // [by virtue of Python's 'auto-reflection' mechanism
        //  we automatically support "double<MYQuantity()" and
        //  others:
        //     __lt__(double,MYQuantity) <==> __gt__(MYQuantity,double)}
        //     ]
        .def(self<double())
        .def(self<=double())
        .def(self==double())
        .def(self!=double())
        .def(self>double())
        .def(self>=double())
        // int
        .def(self<int())
        .def(self<=int())
        .def(self==int())
        .def(self!=int())
        .def(self>int())
        .def(self>=int())
        // and for convenience, string
        .def(self<string())
        .def(self<=string())
        .def(self==string())
        .def(self!=string())
        .def(self>string())
        .def(self>=string())
    ;
    // MVTime
    scope inmytime = class_<MYTime>("mvtime")
        .def(init<>())
        .def(init<const MYTime&>())
        .def(init<const MYQuantity&>())
        .def(init<const string&>())
        .def("__str__", &MYTime::str)
        .def("__deepcopy__", &MYTime::deepcopy)
        // both format() methods
        .def("format", &MYTime::format, mytimeformat())
        // both get() methods
        .def("get", &MYTime::get, mytimeget())
        // Arithmetic
        .def(self+self)
        .def(self+=self)
        .def(self+OQuant())
        .def(self+=OQuant())
        .def(self+string())
        .def(self+=string())
        .def(self-self)
        .def(self-=self)
        .def(self-OQuant())
        .def(self-=OQuant())
        .def(self-string())
        .def(self-=string())
        // support string[+-]MYTime and Quantity[+-]Time
        .def("__radd__", &MYTime::operator+<string>)
        .def("__radd__", &MYTime::operator+<MYQuantity>)
        .def("__rsub__", &MYTime::operator-<string>)
        .def("__rsub__", &MYTime::operator-<MYQuantity>)
        // Comparison
        // to own type
        .def(self<self)
        .def(self<=self)
        .def(self==self)
        .def(self!=self)
        .def(self>self)
        .def(self>=self)
        // to string
        .def(self<string())
        .def(self<=string())
        .def(self==string())
        .def(self!=string())
        .def(self>string())
        .def(self>=string())
        // To quantity
        .def(self<OQuant())
        .def(self<=OQuant())
        .def(self==OQuant())
        .def(self!=OQuant())
        .def(self>OQuant())
        .def(self>=OQuant())
        // the default printing format
        .def_readwrite("strformat",&MYTime::defStrFormat)
    ;
    enum_<MVTime::formatTypes>("formatTypes")
        .value("ANGLE", MVTime::ANGLE)
        .value("TIME", MVTime::TIME)
        .value("CLEAN", MVTime::CLEAN)
        .value("YMD", MVTime::YMD)
        .value("DMY", MVTime::DMY)
        .value("MJD", MVTime::MJD)
        .value("FITS", MVTime::FITS)
        .value("DAY", MVTime::DAY)
    ;
}
